# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""Process a data set such as converting specific columns to numeric values, imputing the missing values, etc.
The module provides several static methods."""

import sys
import re
import multiprocessing
from multiprocessing import get_context
from functools import partial
import warnings
import logging
import pandas as pd
import numpy as np
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.preprocessing import MultiLabelBinarizer
from nltk.tokenize.treebank import TreebankWordDetokenizer, TreebankWordTokenizer
from nltk.stem.wordnet import WordNetLemmatizer
from nltk.stem.arlstem import ARLSTem
from nltk.stem.cistem import Cistem
from nltk.stem.isri import ISRIStemmer
from nltk.stem.lancaster import LancasterStemmer
from nltk.stem.porter import PorterStemmer
from nltk.stem.regexp import RegexpStemmer
from nltk.stem.rslp import RSLPStemmer
from nltk.stem.snowball import SnowballStemmer

from learner.utilities.exclude import exclude
from learner.configuration.configuration import Configuration


class DataProcessor(object):
    def __init__(self, conf: Configuration, exclude_cols=None):
        """Initialize a DataProcessor object.

        :param conf: an instance of Configuration class, which holds all configuration parameters
        :param exclude_cols: a list of column to be excluded from processing.
        """
        self._conf = conf
        self._dummy_encoder = None
        self._label_encoder = None
        self._learner_fill_nan = None
        self._standard_scaler = None
        self._min_max_scaler = None
        self._mean_dict = {}
        self._exclude_cols = exclude_cols if exclude_cols else []
        self._col_to_tfidf_object = {}
        self._col_to_tfidf_object = {}

    @property
    def conf(self):
        return self._conf

    @property
    def dummy_encoder(self):
        return self._dummy_encoder

    @property
    def label_encoder(self):
        return self._label_encoder

    @property
    def learner_fill_nan(self):
        return self._learner_fill_nan

    @property
    def standard_scaler(self):
        return self._standard_scaler

    @property
    def min_max_scaler(self):
        return self._min_max_scaler

    @property
    def exclude_cols(self):
        return self._exclude_cols

    @exclude_cols.setter
    def exclude_cols(self, value):
        self._exclude_cols = value

    @property
    def mean_dict(self):
        return self._mean_dict

    @mean_dict.setter
    def mean_dict(self, value):
        self._mean_dict = value

    @staticmethod
    def handle_copy_cols(data, cols=None):
        """Accept a dictionary with column names and their corresponding names for copying and make a copy of them.
        This method is useful in situations that data processing may change a column but we may still need the original
        column without any changes.

        :param data: the data frame
        :param cols: a list of columns to be copied
        :return: None
        """
        for col, copy_name in cols.items():
            logging.info(f"Copying {col} to {copy_name}")
            data[copy_name] = data[col]

    @exclude(name="cols")
    def make_tuple(self, data, cols=None, exclude_list=None):
        """Turn a series of character strings containing (comma, pipe, caret: ",", "|", "^") separated integers into
        a tuple

        :param cols: column identified by a list of column names
        :param data: the data frame
        :param exclude_list: a list of items that should be excluded from cols (only the decorator function uses this)
        :return: a data frame with the column identified previously converted into a series of tuples
        """
        for col in cols:
            logging.info("Tuplizing column %s", col)
            if isinstance(data[col].iloc[0], tuple):
                logging.info("Column %s has already been tuplized", col)
            else:
                try:
                    # note that sorting is performed to speed up the process of extension later
                    data[col] = data[col].apply(
                        lambda x: tuple(sorted(list(map(int, re.split("[',','|','^']", x))))))
                except TypeError:
                    logging.exception("Column passed '%s' can't be tupilized.", col)
                    raise
        return data

    def handle_dummies(self, data, cols=None):
        """Convert categorical variables into dummy/indicator variables.

        :param data: pandas DataFrame
        :param cols: a list of columns to transform
        :return: an updated pandas DataFrame
        """
        if not self._dummy_encoder:
            self._dummy_encoder = LearnerDummyEncoder()
            return self._dummy_encoder.fit_transform(data, cols)
        else:
            return self._dummy_encoder.transform(data)

    @exclude(name="cols")
    def handle_to_numeric(self, data, cols=None, exclude_list=None):
        """Wrapper around pandas to_numeric method to convert selected columns to a numeric type.

        :param data: a pandas dataframe
        :param cols: the columns to be used for conversion. Learner uses the cat_cols defined by users for conversion.
        :param exclude_list: a list of items that should be excluded from cols (only the decorator function uses this)
        :return: updated data frame
        """
        # first check to see any num_cols is defined, if not issue a warning and return the data
        if cols:
            data[cols] = data[cols].apply(pd.to_numeric, errors='coerce')
            return data
        warnings.warn("No columns were passed to convert to numeric datatype.", UserWarning)
        return data

    @exclude(name="cols")
    def handle_standard_scaler(self, data, cols=None, exclude_list=None):
        """Accept a pandas dataframe and list of columns to standardized those columns by removing the mean and scaling
        to unit variance, i.e. scaling the data so that the mean of the column is 0 and the standard deviation is 1.
        If the standard_scaler attribute is already populated, we use it to transform the data. This is usually the
        case when using the test data. If standard_scaler attribute is not populated, we first fit it to the data and
        then transform it. This is usually the case when training the models.

        :param data: a pandas dataframe
        :param cols: a list of column to use for scaling
        :param exclude_list: a list of items that should be excluded from cols (only the decorator function uses this)
        :return: the updated data
        """
        if cols:
            try:
                if not self.standard_scaler:
                    # try to do the scaling in-place
                    self._standard_scaler = StandardScaler(copy=False)
                    self._standard_scaler.fit(data[cols])
                    data[cols] = self._standard_scaler.transform(data[cols])
                else:
                    data[cols] = self._standard_scaler.transform(data[cols])
            except Exception as e:
                logging.critical(f"Unable to perform standard scaler for the selected columns. "
                                 f"The error is {str(e)}. Exiting...")
                sys.exit(1)
        return data

    @exclude(name="cols")
    def handle_min_max_scaler(self, data, cols=None, exclude_list=None):
        """Accept a pandas dataframe and list of columns to scale those columns between 0 and 1.
        If the min_max_scaler attribute is already populated, we use it to transform the data. This is usually the
        case when using the test data. If min_max_scaler attribute is not populated, we first fit it to the data and
        then transform it. This is usually the case when training the models.

        :param data: a pandas dataframe
        :param cols: a list of column to use for scaling
        :param exclude_list: a list of items that should be excluded from cols (only the decorator function uses this)
        :return: the updated data
        """
        if cols:
            try:
                if not self.min_max_scaler:
                    # try to do the scaling in-place
                    self._min_max_scaler = MinMaxScaler(copy=False)
                    self._min_max_scaler.fit(data[cols])
                    data[cols] = self._min_max_scaler.transform(data[cols])
                else:
                    data[cols] = self._min_max_scaler.transform(data[cols])
            except Exception as e:
                logging.critical(f"Unable to perform min_max scaler for the selected columns. "
                                 f"The error is {str(e)}. Exiting...")
                sys.exit(1)
        return data

    @exclude(name="cols")
    def handle_label_encoding(self, data, cols=None, exclude_list=None):
        """label encode selected columns.

        :param data: pandas dataframe
        :param cols: the columns to use for label encoding
        :param exclude_list: a list of items that should be excluded from cols (the decorator function uses this)
        :return: None
        """
        if not self._label_encoder:
            self._label_encoder = LearnerLabelEncoder()
            self._label_encoder.fit_transform(data, cols)

        else:
            self._label_encoder.transform(data, cols)

    def fill_nan(self, data, default_value, mean_cols, median_cols, mode_cols, value_cols):
        """Fill the missing values using LearnerFillNaN class. If learner_fill_nan is already populated, we only do
        transforms otherwise we do both fit and transform. Fit & transform usually happens during training and transform
        typically happens during testing.

        :param data: pandas dataframe
        :param default_value: a default number to use for filling missing value if no custom value is provided.
        :param mean_cols: a list of columns to be imputed using the mean of the training data
        :param median_cols: a list of columns to be imputed using the median of the training data
        :param mode_cols: a list of columns to be imputed using the mode of the training data
        :param value_cols: a dictionary that maps the column names with a value for imputation
        :return: None
        """
        if not self._learner_fill_nan:
            self._learner_fill_nan = LearnerFillNaN(default_value=default_value,
                                                    mean_cols=mean_cols,
                                                    median_cols=median_cols,
                                                    mode_cols=mode_cols,
                                                    value_cols=value_cols)
            self._learner_fill_nan.fit_transform(data)
        else:
            self._learner_fill_nan.transform(data)

    @staticmethod
    def drop_cols(data, cols=None, **kwargs):
        """Wrapper around pandas drop method for inplace dropping of certain columns

        :param data: pandas dataframe
        :param cols: the columns to be dropped
        :param kwargs: other arguments to pass to drop method such as errors, etc
        :return: None
        """
        if cols:
            data.drop(cols, axis=1, inplace=True, **kwargs)

    def date_cols(self, data, cols=None):
        """Process the date columns. This method, decides whether we should use multiprocessing or not. It then calls
        the appropriate methods. The method _handle_date_cols does the actual work.

        :param data: pandas dataframe
        :param cols: the date columns to be processed
        :return: the updated dataframe with the new columns
        """
        if cols:
            if self.conf.process.date_cols_num_cores == 1 or self.conf.process.date_cols_num_cores is None:
                data = handle_date_cols(data, cols)
            else:
                data = parrallelize_dataframe(data, partial(handle_date_cols, cols=cols),
                                              num_cores=self.conf.process.date_cols_num_cores)

            return data

    def handle_mean(self, data):
        """Accept a dataframe and compute the mean of all columns in that dataframe and save them in a dictionary (mean_dict)
        in which the key is the column name and the value is the average value of that column.

        :param data: pandas dataframe
        :return: None
        """
        self.mean_dict = data.mean().to_dict()

    @staticmethod
    def multilabel_binarizer(data, target_col, other_cols=None):
        """Use pandas MultiLabelBinarizer to binarize a certain columns. The return value is a dataframe with the
        columns set using classes\_ instance member of MultiLabelBinarizer joined with any other columns (other_cols)
        asked by the user

        :param data: pandas dataframe
        :param target_col: the column to perform MultiLabelBinarizer on
        :param other_cols: other columns in the dataset that the return dataset should contain
        :return: pandas dataframe and performing binarization and pandas MultiLabelBinarizer object
        """
        # Note: A check needs to be done to make sure other_cols actually exists in the data set
        mlb = MultiLabelBinarizer(sparse_output=False)
        df = pd.DataFrame(mlb.fit_transform(data[target_col]), columns=mlb.classes_)
        if other_cols:
            return df.join(data[other_cols]), mlb
        return df, mlb

    def handle_lemmatize(self, data, params):
        """Accept a pandas dataframe and a list of dictionaries (params) to perform the text lemmatization on the
        selected columns. We run the text lemmatization is parallel. The method _perform_lemmatization handles the main
        operations for text lemmatizetion. Here, we loop through the list of pos.

        :param data: a pandas dataframe, it could be train, validation, or test data
        :param params: a list of dictionaries containing the information about the columns and list of "pos"
        :return: an updated dataframe with the lemmatized form of text columns
        """
        for param in params:
            for pos in param["pos"]:
                logging.info(f"Performing lemmatization for {param['name']} and pos={pos}")
                data = parrallelize_dataframe(data, partial(self._perform_lemmatization,
                                                            col=param["name"],
                                                            pos=pos))
        return data

    def _perform_lemmatization(self, data, col, pos):
        """This method handles the heavy lifting for the text lemmatization. For each row, it first tokenizes the text.
        Then, it lemmatizes each token in the text. Last, it detokenizes the tokens and returns the updated data.

        :param data: a pandas dataframe, it could be train, validation, or test data
        :param col: the name of the column to be lemmatized
        :param pos: the part of speech
        :return: an updated dataframe with the lemmatized form of text columns
        """
        tokenizer = TreebankWordTokenizer()
        detokenizer = TreebankWordDetokenizer()
        lemmatizer = WordNetLemmatizer()
        # here we tokenize the text, lemmatize it, and then detokenize it
        data[col] = data[col].apply(lambda x: detokenizer.detokenize(
            [lemmatizer.lemmatize(word, pos=pos) for word in tokenizer.tokenize(x)]
        ))
        return data

    def handle_stem(self, data, params):
        """Accept a pandas dataframe and a list of dictionaries (params) to perform the text stemming on the
        selected columns. We run the text stemming is parallel. The method _perform_stemming handles the main
        operations for text stemming. Here, we loop through the list of pos.

        :param data: a pandas dataframe, it could be train, validation, or test data
        :param params: a list of dictionaries containing the information about the columns, the type of semmer, and additional options
        :return: an updated dataframe with the stemmed form of text columns
        """
        for param in params:
            logging.info(f"Performing stemming for {param['name']}")
            data = parrallelize_dataframe(data, partial(self._perform_stemming,
                                                        col=param["name"],
                                                        type_=param["type"],
                                                        options=param["options"]))
        return data

    def _perform_stemming(self, data, col, type_, options):
        """This method handles the heavy lifting for the text stemming. For each row, it first tokenizes the text.
        Then, it stems each token in the text. Last, it detokenizes the tokens and returns the updated data.

        :param data: a pandas dataframe, it could be train, validation, or test data
        :param col: the name of the column to be stemmed
        :param type_: the type of stemmer
        :param options: additional options for the stemmer. This is required for some stemmers.
        :return: an updated dataframe with the stemmed form of text columns
        """
        tokenizer = TreebankWordTokenizer()
        detokenizer = TreebankWordDetokenizer()
        stemmer = getattr(sys.modules[__name__], type_)(**options)

        # here we tokenize the text, stem it, and then detokenize it
        data[col] = data[col].apply(lambda x: detokenizer.detokenize(
            [stemmer.stem(word) for word in tokenizer.tokenize(x)]
        ))
        return data

    def handle_count_vectorize(self, data, params):
        """Accept a pandas dataframe and a list of dictionaries (params) to perform the count vectorize operations on the
        selected columns. We fit and transform during the training and only transform during testing or validation. We
        get to know if we are training or testing based on the _col_to_count_vectorize_object dictionary. If this
        dictionary is None, that means we are training, otherwise we are testing/validating.

        :param data: a pandas dataframe, it could be train, validation, or test data
        :param params: a list of dictionaries containing the information about the column and bag_of_word operation
        :return: an updated dataframe with new columns for the term frequency matrix
        """
        # if _col_to_count_vectorize_object is None, we need to train
        if not self._col_to_tfidf_object:
            for param in params:
                count_vector = CountVectorizer(strip_accents=param["strip_accents"],
                                               lowercase=param["lowercase"],
                                               stop_words=param["stop_words"],
                                               token_pattern=param["token_pattern"],
                                               ngram_range=param["ngram_range"],
                                               analyzer=param["analyzer"],
                                               max_df=param["max_df"],
                                               min_df=param["min_df"],
                                               max_features=param["max_features"],
                                               binary=param["binary"])
                doc_array = count_vector.fit_transform(data[param["name"]])
                frequency_df = pd.DataFrame.sparse.from_spmatrix(data=doc_array, columns=count_vector.get_feature_names()).reset_index(drop=True)
                data = pd.concat([frequency_df, data], axis=1)
                # add the object to the dictionary to use it during testing
                self._col_to_tfidf_object[param["name"]] = count_vector
                # we need to drop the column because the training algorithms won't accept it
                self._conf.column.drop_cols.append(param["name"])
        else:
            for col, obj in self._col_to_tfidf_object.items():
                # use the pre-trained object to do the transform
                doc_array = obj.transform(data[col])
                frequency_df = pd.DataFrame.sparse.from_spmatrix(data=doc_array, columns=obj.get_feature_names())
                data = pd.concat([frequency_df, data], axis=1)
                self._conf.column.drop_cols.append(col)
        return data

    def handle_tfidf(self, data, params):
        """Accept a pandas dataframe and a list of dictionaries (params) to perform the
        tf-idf (term frequency - inverse document frequency)  operations on the
        selected columns. We fit and transform during the training and only transform during testing or validation. We
        get to know if we are training or testing based on the _col_to_tfidf_object dictionary. If this
        dictionary is None, that means we are training, otherwise we are testing/validating.

        :param data: a pandas dataframe, it could be train, validation, or test data
        :param params: a list of dictionaries containing the information about the column and bag_of_word operation
        :return: an updated dataframe with new columns for the term frequency matrix
        """
        # if _col_to_tfidf_object is None, we need to train
        if not self._col_to_tfidf_object:
            for param in params:
                tfidf_vector = TfidfVectorizer(strip_accents=param["strip_accents"],
                                               lowercase=param["lowercase"],
                                               stop_words=param["stop_words"],
                                               token_pattern=param["token_pattern"],
                                               ngram_range=param["ngram_range"],
                                               analyzer=param["analyzer"],
                                               max_df=param["max_df"],
                                               min_df=param["min_df"],
                                               max_features=param["max_features"],
                                               binary=param["binary"],
                                               norm=param["norm"],
                                               use_idf=param["use_idf"],
                                               smooth_idf=param["smooth_idf"],
                                               sublinear_tf=param["sublinear_tf"])
                doc_array = tfidf_vector.fit_transform(data[param["name"]])
                frequency_df = pd.DataFrame.sparse.from_spmatrix(data=doc_array, columns=tfidf_vector.get_feature_names()).reset_index(drop=True)
                data = pd.concat([frequency_df, data], axis=1)
                # add the object to the dictionary to use it during testing
                self._col_to_tfidf_object[param["name"]] = tfidf_vector
                # we need to drop the column because the training algorithms won't accept it
                self._conf.column.drop_cols.append(param["name"])
        else:
            for col, obj in self._col_to_tfidf_object.items():
                # use the pre-trained object to do the transform
                doc_array = obj.transform(data[col])
                frequency_df = pd.DataFrame.sparse.from_spmatrix(data=doc_array, columns=obj.get_feature_names())
                data = pd.concat([frequency_df, data], axis=1)
                self._conf.column.drop_cols.append(col)
        return data

    def memory_optimization(self, data):
        """Reduce memory usage by converting dtypes down to lowest bit size.

        There is a known bug existing in pandas where it will downcast float64 to float32. This should not have an
        impact at this time, but will need to be accounted for in the unit test.

        :param data: pandas dataframe
        :return: pandas dataframe with reduced bit scope for int columns.
        """
        logging.info("Memory Usage before optimization: %s", self.mem_usage(data))

        for col in data.columns:
            if str(data[col].dtypes).startswith("uint"):
                data[col] = pd.to_numeric(data[col], downcast='unsigned')
            if str(data[col].dtypes).startswith("int"):
                data[col] = pd.to_numeric(data[col], downcast='integer')
            if str(data[col].dtypes).startswith("float"):
                data[col] = pd.to_numeric(data[col], downcast='float')

        logging.info("Memory Usage after optimization: %s", self.mem_usage(data))
        if self._conf.workspace.verbose_level == "DEBUG":
            for col in data.columns:
                logging.info("%s: %s", col, data[col].dtype)

    @staticmethod
    def mem_usage(data):
        """Returns human readable MB usage of Dataframe

        :param data: pandas dataframe
        :return: Memory usage in MB
        """
        return "{:03.2f}MB".format(data.memory_usage().sum() / 1024 ** 2)

    @exclude(name="cols")
    def log_transform(self, data, params, cols=None, exclude_list=None):
        """Use the data and params to call the log_transform method in feature_engineering module. This is currently used
        for log transformation of the target column.

        :param data: a pandas dataframe
        :param params: a list of log_transform parameters coming from the conf object (see the configuration module for details)
        :param cols: a list of columns for log transformation. This is provided so that exclude_list can be applied to it.
        :param exclude_list: a list of items that should be excluded from cols (only the decorator function uses this)
        :return: updated data frame
        """
        if cols:
            from learner.feature_engineering.feature_engineering import FeatureEngineering
            data = FeatureEngineering.log_transform(data, params)
            # in case, we have NaN after the transformation, we use the DataValidator method to drop them
            from learner.validator.data_validator import DataValidator
            DataValidator.check_nulls_in_col(data, col=cols[0])
        return data

    @staticmethod
    def exponential_transform(data, params, cols=None):
        """Use the data and params to call the exponential_transform method in feature_engineering module. This is currently used
        for to transform back the log transformed target.

        :param data: a pandas dataframe
        :param params: a list of exponential_transform parameters coming from the conf object (see the configuration module for details)
        :param cols: a list of columns for log transformation. This is provided so that exclude_list can be applied to it.
        :return: updated data frame
        """
        if cols:
            from learner.feature_engineering.feature_engineering import FeatureEngineering
            data = FeatureEngineering.exponential_transform(data, params)
        return data

    def process_data(self, data):
        """The main function for data processing. Depending on the user's input the appropriate if blocks will execute.

        :return: conf, data, and segmenter objects
        """

        logging.info("Processing the data")

        if self._conf.column.copy_cols:
            self.handle_copy_cols(data, cols=self._conf.column.copy_cols)

        if self._conf.analysis.shap_activate and not self.mean_dict and self.conf.analysis.shap_use_training_mean:
            logging.info("Calculating the mean of each feature in training data to use in narratives...")
            self.handle_mean(data)

        if self._conf.process.tuplize_activate and self._conf.process.tuplize_cols:
            self.make_tuple(data, cols=self._conf.process.tuplize_cols, exclude_list=self._exclude_cols)

        if self._conf.process.to_numeric_activate and self._conf.process.to_numeric_cols:
            logging.info("Handling to_numeric...")
            data = self.handle_to_numeric(data, cols=self._conf.process.to_numeric_cols, exclude_list=self._exclude_cols)

        if self._conf.process.standard_scaler_activate and self._conf.process.standard_scaler_cols:
            logging.info("Handling standard scaling... ")
            data = self.handle_standard_scaler(data, cols=self._conf.process.standard_scaler_cols, exclude_list=self._exclude_cols)

        if self._conf.process.min_max_scaler_activate and self._conf.process.min_max_scaler_cols:
            logging.info("Handling min_max scaling... ")
            data = self.handle_min_max_scaler(data, cols=self._conf.process.min_max_scaler_cols, exclude_list=self._exclude_cols)

        if self._conf.process.date_cols_activate:
            logging.info("Handling date columns...")
            data = self.date_cols(data, self._conf.process.date_cols_params)

        if self._conf.process.fillnan_activate:
            logging.info("Filling missing values in the dataset...")
            self.fill_nan(data,
                          default_value=self._conf.process.fillnan_value,
                          mean_cols=self._conf.process.fillnan_mean_cols,
                          median_cols=self._conf.process.fillnan_median_cols,
                          mode_cols=self._conf.process.fillnan_mode_cols,
                          value_cols=self._conf.process.fillnan_value_cols)

        if self._conf.process.label_encoding_activate and self._conf.process.label_encoding_cols:
            logging.info("Label encoding selected columns...")
            self.handle_label_encoding(data, cols=self._conf.process.label_encoding_cols, exclude_list=self._exclude_cols)

        if self._conf.process.dummies_activate and self._conf.process.dummies_cols:
            logging.info("Handling to_dummies...")
            data = self.handle_dummies(data, cols=self._conf.process.dummies_cols)

        if self.conf.process.lemmatize_activate:
            logging.info("Lemmatizing the selected columns...")
            data = self.handle_lemmatize(data, params=self._conf.process.lemmatize_cols_params)

        if self.conf.process.stem_activate:
            logging.info("Stemming the selected columns...")
            data = self.handle_stem(data, params=self._conf.process.stem_cols_params)

        if self._conf.process.count_vectorize_activate:
            logging.info("Performing count vectorize operations for the selected columns...")
            data = self.handle_count_vectorize(data, params=self._conf.process.count_vectorize_cols_params)

        if self._conf.process.tfidf_activate:
            logging.info("Performing tfidf operations for the selected columns...")
            data = self.handle_tfidf(data, params=self._conf.process.tfidf_cols_params)

        if self._conf.column.drop_cols:
            logging.info("Dropping the selected columns...")
            self.drop_cols(data, self._conf.column.drop_cols)

        if self._conf.process.memory_optimization:
            logging.info("Running memory optimization...")
            self.memory_optimization(data)

        if self._conf.process.log_transform_target_activate:
            logging.info("Log transforming the target column...")
            self.log_transform(data, params=self._conf.process.log_transform_target_params, cols=[self._conf.column.target_col], exclude_list=self._exclude_cols)

        # set this so that the target is excluded when making predictions
        self._exclude_cols = [self._conf.column.target_col]

        logging.info("Successfully processed the data")
        return self._conf, data


class LearnerDummyEncoder:
    """LearnerDummyEncoder is a customized encoder that takes into consideration instances where the test data
    may include labels that are not present in the training data or the instances where the test data is missing some
    levels that were present in the training data. In such cases, LearnerDummyEncoder ensures that features in the training
    and testing datasets are identical. Note that LearnerDummyEncoder can accept categorical integer features as well as
    other types of categorical features."""

    def __init__(self):
        """Initialize a LearnerDummyEncoder by assigning the necessary instance attributes."""
        self._cols = None
        self._classes = set()
        self._all_cols = []

    @property
    def cols(self):
        return self._cols

    @property
    def classes(self):
        return self._classes

    @property
    def all_cols(self):
        return self._all_cols

    def fit(self, data, cols):
        """Generate classes (new column names after transformation) and transform data.

        :param data: pandas DataFrame
        :param cols: a list of columns for which to perform dummy encoding
        :return data: the updated DaraFrame
        """
        self._cols = cols
        data, self._classes, self._all_cols = self._get_classes(data, self._cols)
        return data

    def transform(self, data):
        """Transform the data using dummy encoding assuming the classes are already populated. Drop columns that were not
        present in the traning data and generate columns that are missing.

        :param data: pandas DataFrame
        :return data: an updated pandas DataFrame
        """
        if self._classes:
            data, classes, _ = self._get_classes(data, self._cols)

            missing_classes = list(self._classes - classes)
            missing_classes.sort(key=lambda x: self._all_cols.index(x))
            for col in missing_classes:
                data.insert(self._all_cols.index(col), col, 0)

            drop_cols = classes - self._classes
            data.drop(drop_cols, axis=1, inplace=True)

            return data

    def fit_transform(self, data, cols):
        """Call the fit method to process the data and get the updated data

        :param data: pandas DataFrame
        :param cols: a list of column to
        :return data: the updated DataFrame
        """
        return self.fit(data, cols)

    @staticmethod
    def _get_classes(data, cols):
        """Call pandas get_dummies method to convert categorical variables into dummy/indicator variables. Then
        create a set that contains the name of the new columns that were created as well as a list that contain all
        the columns.

        :param data: a pandas DataFrame
        :param cols: a list of columns to be converted into dummy variables
        :return: the updated DataFrame, a set containing the new columns, and a list of all columns
        """
        data = pd.get_dummies(data, prefix_sep='___', columns=cols, dummy_na=True)

        classes = set()
        columns = data.columns
        for col in cols:
            classes.update(columns[columns.str.contains(col + "___")])

        return data, classes, list(columns)


class LearnerLabelEncoder:
    """LearnerLabelEncoder is a customized label encoder that takes into consideration instances where the test data
    may include labels that are not present in the training data. In such case, new labels will be coded and added
    to a dictionary that includes all labels and their codes. This will maintain consistency in the way training
    and testing datasets are being encoded."""

    def __init__(self, encode_unique_classes=False):
        """Initialize LearnerLabelEncoder using a flag which determines how the new levels should be encoded/handled

        :param encode_unique_classes: if set to True, each unique class will have a unique code. If set to False
        all classes that don't exist in training data will be assigned a label of -1
        """
        self._labels_dict = dict()
        self._encode_unique_classes = encode_unique_classes

    @property
    def labels_dict(self):
        return self._labels_dict

    @property
    def encode_unique_classes(self):
        return self._encode_unique_classes

    def fit(self, data, cols):
        """Creates a dictionary of labels and their corresponding classes for a set of columns.

        :param data: data in pandas dataframe format
        :param cols: columns to be label encoded
        :return: None
        """
        for col in cols:
            try:
                classes = sorted(data[col].unique())
            except TypeError:
                # if we fail with the TypeError, we won't sort
                classes = data[col].unique()
            except Exception:
                # in we fail again, convert to string and sort. We do not expect to get here
                classes = sorted(data[col].astype(str).unique())
            self._labels_dict[col] = {class_value: label for label, class_value in enumerate(classes)}

    def transform(self, data, cols):
        """If LearnerLabelEncoder is already fit, checks if the test data has any labels that are not in training. If
        this is the case, update self._labels_dict (labels dictionary) to include the new label and its code. Then,
        perform transformation based on the values of labels_dict.

        :param data: data in pandas dataframe
        :param cols: columns to do transformation on
        :return: None
        """
        if self._labels_dict:
            for col in cols:
                test_classes = data[col].unique()
                diff = set(test_classes) - set(self._labels_dict[col].keys())
                max_label = max(self._labels_dict[col].values())
                if diff:
                    for new_class in diff:
                        if self._encode_unique_classes:
                            new_label = max(self._labels_dict[col].values()) + 1
                            self._labels_dict[col][new_class] = new_label
                        else:
                            warnings.warn(f"Label {new_class} in column {col} is not in training. Encoding this new "
                                          f"labels with {max_label + 1}...", Warning)
                            self._labels_dict[col][new_class] = max_label + 1

                data[col] = data[col].apply(self._labels_dict[col].get)
                # data[col] = data[col].apply(lambda x: self._labels_dict[col][x])

    def fit_transform(self, data, cols):
        """Fit and transform columns using Learner LabelEncoder.

        :param data: data in pandas dataframe
        :param cols: columns to fit and transform
        :return: None
        """
        logging.info("Fitting data to Learner LabelEncoder...")
        self.fit(data, cols)
        logging.info("Transforming data using Learner LabelEncoder...")
        self.transform(data, cols)

    def inverse_transform(self, y, col):
        """Transform labels back to their original classes

        :param y: single column (pandas series or numpy array)
        :param col: column name to be transformed (used to pull the dictionary of encoded values from self._labels_dict)
        :return: a series of transformed labels
        """
        logging.info("Inverse transformation of labels in column '%s' to original values using Learner LabelEncoder...",
                     col)
        inverse_dictionary = {val: key for (key, val) in self._labels_dict[col].items()}
        y_original = pd.Series(y).map(inverse_dictionary)
        return y_original.values


class LearnerFillNaN:
    """LearnerFillNaN is a class for handling imputation of missing values in datasets. The reason for implementing a
     custom class is to be able to use the values from the training data when imputing the test data. For example, if
     the missing values were imputed using the "mean" of the training data, we'd want to use the same values when
     imputing the test data instead of using the mean of testing data.
     """

    def __init__(self, default_value, mean_cols, median_cols, mode_cols, value_cols):
        """Initialize LeanerFillNaN using the default_value for imputation and a dictionary that maps the column names
        to the value/method for imputation.

        :param default_value: a default number to use for filling missing value if no custom value is provided.
        :param mean_cols: a list of columns to be imputed using the mean of the training data
        :param median_cols: a list of columns to be imputed using the median of the training data
        :param mode_cols: a list of columns to be imputed using the mode of the training data
        :param value_cols: a dictionary that maps the column names with a value for imputation
        """
        self._mean_dict = dict()
        self._median_dict = dict()
        self._mode_dict = dict()
        self._default_value = default_value
        self._mean_cols = mean_cols
        self._median_cols = median_cols
        self._mode_cols = mode_cols
        self._value_cols = value_cols

    @property
    def mean_dict(self):
        return self._mean_dict

    @property
    def median_dict(self):
        return self._median_dict

    @property
    def mode_dict(self):
        return self._mode_dict

    @property
    def mean_cols(self):
        return self._mean_cols

    @property
    def median_cols(self):
        return self._median_cols

    @property
    def mode_cols(self):
        return self._mode_cols

    @property
    def default_value(self):
        return self._default_value

    def fit(self, data):
        """Loop through fillnan_cols to populate the fillnan_dict. fillnan_dict maps the column names to the
        corresponding numbers for imputation of the missing values. If the value in the fillnan_col dictionary is a
        string, we call the _get_fill_value_for_method to compute the mean, median, or mode.

        :param data: pandas dataframe
        :return: None
        """
        for col in self.mean_cols:
            self.mean_dict[col] = self._get_fill_value_for_method(data,
                                                                  col,
                                                                  method="mean",
                                                                  default_value=self.default_value)
        for col in self.median_cols:
            self.median_dict[col] = self._get_fill_value_for_method(data,
                                                                    col,
                                                                    method="median",
                                                                    default_value=self.default_value)
        for col in self.mode_cols:
            self.mode_dict[col] = self._get_fill_value_for_method(data,
                                                                  col,
                                                                  method="mode",
                                                                  default_value=self.default_value)

    @staticmethod
    def _get_fill_value_for_method(data, col, method, default_value):
        """Depending on the method, i.e. "mean", "median", or "mode", compute the values that should be used for
        imputation of the column. If the value cannot be calculated, use the default value and issues a warning.

        :param data: a pandas dataframe
        :param col: the name of the column in the data
        :param method: the imputation method
        :param default_value: the default value to use in case the calculations fail
        :return: the computed value or the default value
        """
        try:
            if method == "mean":
                fill_value = data[col].mean()
            elif method == "median":
                fill_value = data[col].median()
            elif method == "mode":
                fill_value = data[col].mode().iloc[0]
        except TypeError:
            warnings.warn(f"""The column '{col}' cannot be imputed by {method}. Imputing the column {col} with 
                           the {default_value}""", Warning)
            fill_value = default_value
        finally:
            return fill_value

    def transform(self, data):
        """After we fit the data or if the fillnan_dict is already populated, loop through the fillnan_dict and fill the
        missing values in each column with their corresponding values. After that, fill all other missing values with
        the default value.

        :param data: a pandas dataframe
        :return: None
        """
        # this may fail if we have SparseArray
        try:
            data.replace([-np.inf, np.inf], np.NaN, inplace=True)
        except Exception:
            pass

        for col, value in self.mean_dict.items():
            logging.info(f"Filling missing value in {col} with {value}")
            data[col].fillna(value, inplace=True)
        for col, value in self.median_dict.items():
            logging.info(f"Filling missing value in {col} with {value}")
            data[col].fillna(value, inplace=True)
        for col, value in self.mode_dict.items():
            logging.info(f"Filling missing value in {col} with {value}")
            data[col].fillna(value, inplace=True)
        for col, value in self._value_cols.items():
            logging.info(f"Filling missing value in {col} with {value}")
            data[col].fillna(value, inplace=True)
        logging.info("Filling missing values in the unspecified columns with %.2f, if any", self._default_value)
        data.fillna(self._default_value, inplace=True)

    def fit_transform(self, data):
        """Call fit and transform method. This method is typically called during training while the transform method is
        called during testing.

        :param data: a pandas dataframe
        :return: None
        """
        self.fit(data)
        self.transform(data)


def handle_label_encoding(data, cols=None):
    """label encode selected columns.

    :param data: pandas dataframe
    :param cols: the columns to use for label encoding
    :return: pandas DataFrame with encoded data and label_encoder object
    """
    label_encoder = LearnerLabelEncoder()
    if cols:
        label_encoder.fit_transform(data, cols)
        return data, label_encoder


def filter_data(data, column, value, criterion):
    """Filter a pandas DataFrame based on a criterion. Due to the internal mechanism of query method, the query may fail.
    In this situation, if the criterion is equality operation, we then catch the error and filter the data using equality
    operator.

    :param data: pandas DataFrame
    :param column: column to filter
    :param value: value used in conditional
    :param criterion: condition to check
    :return: a filtered dataframe with the original indices

    Example:

        df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})

        filter_data(data=df, column='a', value=2, criterion= '>=')

        output >>

        pd.DataFrame({'a': [2, 3], 'b': [5, 6]})
    """
    try:
        idx = data.query('{0} {1} {2}'.format(column, criterion, value)).index
    except Exception as e:
        # in case things go wrong in equality operator, filter using regular equality operator.
        if criterion == '==':
            logging.error("Filtering the value {value} failed. The error is {error}. Using a "
                          "different method...".format(value=value, error=str(e)))
            idx = data[data[column] == value].index
            logging.info("Successfully filtered the data")
        else:
            raise e
    return data.loc[idx]


def parrallelize_dataframe(df, func, num_cores=None):
    """Accept a dataframe and a function, break the dataframe into multiple pieces (based on the number of cores). Then
    distribute the pieces and call the function on that each piece. In the end, get the result from each processor and
    concatenate the pieces together to return it.

    :param df: a pandas dataframe
    :param func: a function to pass the dataframe to
    :param num_cores: the number of cores to use
    :return: the updated dataframe
    """
    if not num_cores:
        num_cores = multiprocessing.cpu_count()

    logging.info(f"Processing data using {num_cores} cores")

    # this line breaks the dataframe and gives a list of dataframe pieces
    df_split = np.array_split(df, num_cores)
    with get_context("spawn").Pool() as pool:
    # with multiprocessing.Pool(num_cores) as pool:
        df = pd.concat(pool.map(func, df_split), sort=False)
    return df


def parrallelize_dataframes(df1, df2, func, num_cores=None):
    """This functions is similar to parrallelize_dataframe but it accepts two dataframe instead of one.

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param func: a function to pass the dataframe to
    :param num_cores: the number of cores to use
    :return: the updated dataframes
    """
    if not num_cores:
        num_cores = multiprocessing.cpu_count()

    logging.info(f"Processing data using {num_cores} cores")

    df1_split = np.array_split(df1, num_cores)
    df2_split = np.array_split(df2, num_cores)

    with get_context("spawn").Pool() as pool:
    # with multiprocessing.Pool(num_cores) as pool:
        df_lists = pool.starmap(func, zip(df1_split, df2_split))

        df1 = pd.concat((df[0] for df in df_lists), sort=False)
        df2 = pd.concat((df[1] for df in df_lists), sort=False)

    return df1, df2


def log_transform(data, params, cols=None):
    """Use the data and params to call the log_transform method in feature_engineering module. This is currently used
    for log transformation of the target column.

    :param data: a pandas dataframe
    :param params: a list of log_transform parameters coming from the conf object (see the configuration module for details)
    :param cols: a list of columns for log transformation. This is provided so that exclude_list can be applied to it.
    :return: updated data frame
    """
    if cols:
        from learner.feature_engineering.feature_engineering import FeatureEngineering
        data = FeatureEngineering.log_transform(data, params)
    return data


def delete_keys_from_dict(dictionary, keys):
    """Accept a dictionary and a list of keys. Delete all the keys from the dictionary.

    :param dictionary: a dictionary to be processed
    :param keys: a list of key to be removed from the dictionary
    :return: None
    """
    if not isinstance(dictionary, dict):
        logging.critical("Wrong input type. Provide a dictionary as the first argument or the 'dictionary' keyword "
                         "argument. Exiting...")
        sys.exit(1)

    if not isinstance(keys, list):
        logging.critical("Wrong input type. Provide a list as the second argument or the 'keys' keyword argument. "
                         "Exiting...")
        sys.exit(1)

    for key in keys:
        dictionary.pop(key)


def handle_date_cols(data, cols=None):
    """The function was moved out of the DataProcessor class because we'll have issue with using custom scoring
    functions and get can't pickle <function> when doing multiprocessing.
    Process the date columns. First, convert each column to datetime object and then extract the requested items
    from them to create the new columns.

    :param data: pandas dataframe
    :param cols: the date columns to be processed
    :return: the updated dataframe with the new columns
    """
    for col, date_items in cols.items():
        data[col] = pd.to_datetime(data[col],
                                   errors='coerce',
                                   infer_datetime_format=True)
        for item in date_items:
            data.loc[:, f"{col}_{item}"] = getattr(data[col].dt, item)

    return data





# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements classes and functions related to moving data around. This would include moving data to/from
GPU, S3, etc."""
import logging
import sys
import os

import torch
import boto3
from pathlib import Path

from learner.configuration.configuration import Configuration


class DataMoverManager:
    """This class handles the operations related to moving the data around. One example include uploading/downloading
    data to/from S3. In the future, we may add some other functionalities such as SFTP, etc. This class does not
    handle the implementation, instead it uses other classes to do the work."""

    def __init__(self, conf: Configuration):
        """Instantiate a DataMoverManager object using a conf object

        :param conf: an instance of the Configuration class
        """
        self._conf = conf
        self.data_mover = S3DataMover()

    @property
    def conf(self):
        return self._conf

    def move_data(self, mode):
        """Call the necessary methods to move data around depending on the mode. Currently, we only move data to and
        from s3 but in the future we may add other options.

        :param mode: the mode for moving data, it could be "download" or "upload"
        :return: None
        """
        if self.conf.workspace.s3_activate and mode == "download":
            self.data_mover.download_folder_from_s3(self.conf.workspace.s3_path, self.conf.workspace.path)
        if self.conf.workspace.s3_activate and mode == "upload":
            self.data_mover.upload_folder_to_s3(self.conf.workspace.path, self.conf.workspace.s3_path)


class S3DataMover:
    """This class implements method that move data to or from S3. This class only implements methods that we currently
    need.
    """
    def __init__(self):
        """Instantiate an S3DataMover object.
        """
        self._s3_client = boto3.client("s3")

    @property
    def client(self):
        return self._s3_client

    def upload_file_to_s3(self, from_file, to_file):
        """Upload a file to a s3 location. This method is used by copy_folder_to_s3 method.

        :param from_file: full path to the local file
        :param to_file:  full path to the s3 location
        :return: None
        """
        bucket, s3_key = split_s3_path(to_file)

        self._s3_client.upload_file(from_file, bucket, s3_key)

    def upload_folder_to_s3(self, from_folder, to_folder, pattern="*"):
        """Upload a local folder recursively to s3. We currently do not upload the empty folders.

        :param from_folder: a local folder
        :param to_folder: a s3 "folder"
        :param pattern: regular expression pattern passed to "glob" to filter the type of files needed.
        :return: None
        """
        logging.info("Uploading data to s3...")
        bucket, s3_key = split_s3_path(to_folder)
        s3_folder = Path(s3_key)
        parent_folder = Path(from_folder)
        for item in parent_folder.glob(pattern):
            relative = item.relative_to(parent_folder)
            target_full_path = "s3://{}/{}".format(bucket, s3_folder.joinpath(relative))
            if item.is_file():
                self.upload_file_to_s3(str(item), target_full_path)
            else:
                self.upload_folder_to_s3(str(item), target_full_path)
        logging.info("Successfully uploaded data to s3.")

    def download_folder_from_s3(self, from_folder, to_folder):
        """Download all the objects inside a "folder" recursively from s3.

        :param from_folder: the full path to the "folder" on s3.
        :param to_folder: the full path to the local folder to download the objects into
        :return: None
        """
        logging.info("Downloading data from s3...")
        bucket, s3_key = split_s3_path(from_folder)
        # if the folder is empty we return
        if 'Contents' not in self._s3_client.list_objects(Bucket=bucket, Prefix=s3_key):
            logging.info("The s3 location is empty.")
            return
        for key in self._s3_client.list_objects(Bucket=bucket, Prefix=s3_key)['Contents']:
            s3_parent_folder = Path(s3_key)
            relative_path = Path(key['Key']).relative_to(s3_parent_folder)
            target_folder = Path(to_folder)
            full_path = target_folder.joinpath(relative_path)
            if key['Key'].endswith('/'):
                if not os.path.exists(full_path):
                    os.makedirs(full_path)
            else:
                directory = full_path.parents[0]
                if not os.path.exists(directory):
                    os.makedirs(directory)
                self._s3_client.download_file(bucket, key['Key'], str(full_path))
        logging.info("Successfully downloaded data from s3.")

    def delete_from_s3(self, s3_path: str):
        """delete all objects in given s3_path

        :param s3_path: location to the s3 object. e.g. s3://my-awesome-bucket/path/to/object
        :return: None
        """
        bucket, s3_key = split_s3_path(s3_path)
        # if the folder is empty we return
        if 'Contents' not in self._s3_client.list_objects(Bucket=bucket, Prefix=s3_key):
            return
        for key in self._s3_client.list_objects(Bucket=bucket, Prefix=s3_key)['Contents']:
            self._s3_client.delete_object(Bucket=bucket, Key=key['Key'])


def move_to_device(data, device):
    """Move the data (a single or a list of PyTorch tensors) to a specific device. The device would be cpu or cuda. We
    currently are not validating the devices because device is validated when instantiating the conf object. The reason
    for implementing this method is that in some cases, we have a list of tensors that need to be moved instead of
    a single tensor. This function can handle those cases as well.

    :param data: the data (mainly a PyTorch tensor or a list of PyTorch tensors)
    :param device: the name of the device, could be "cpu" or "cuda"
    :return: the same data in that device.
    """
    if isinstance(data, torch.Tensor):
        return data.to(device)
    if isinstance(data, list):
        for i in range(len(data)):
            data[i] = data[i].to(device)
        return data


def split_s3_path(s3_path):
    """Split the s3_path into s3 key and bucket. The result will all be lower case.

    Example:
    >>> bucket, s3_key = split_s3_path("s3://bucket/path/to/object")
    >>> bucket
    'bucket'
    >>> s3_key
    'path/to/object'

    :param s3_path: a full s3 path to an object, e.g. s3://bucket/path/to/object
    :return: a tuple of strings, (bucket, s3_key)
    """
    if not s3_path.startswith("s3://"):
        logging.critical("The S3 path should start with 's3://'. Exiting...")
        sys.exit(1)

    s3_path = s3_path[5:]
    bucket, s3_key = s3_path.split("/", 1)
    return bucket, s3_key



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""logger is responsible for preparing the logging configuration. The Logger module parses the relevant information
from the configuration file and updates the default values of the logging configuration."""

import sys
import os
import json
import commentjson
import warnings
import logging.config


class Logger:
    """Get the sep_timetag and configure the logging settings using the information passed through the configuration
     file."""
    def __init__(self, json_config, sep_timetag):
        """Instantiate the Logger object using a sep_timetag and a jaosn configuration file.

        :param json_config: the path to the configuration file
        :param sep_timetag: the sep + timetag
        """

        self._json_config = json_config
        self.sep_timetag = sep_timetag
        self.output_path = self.get_output_path()
        self.output_name = self.get_output_name()
        self.create_log_file = self.get_create_log_file()
        self.verbose_level = self.get_verbose_level()
        self.setup_logging(os.path.join(os.path.dirname(__file__),
                                        "..",
                                        "schema",
                                        "logging.json"))

    @property
    def json_config(self):
        return self._json_config

    def get_output_path(self):
        try:
            path = self._json_config["workspace"]["path"]
            assert isinstance(path, str), "path must be a string"
            if path:
                if os.path.isdir(path) is False:
                    warnings.warn("The path provided is not an existing directory, attempting to create the  "
                                  "directory...")
                    try:
                        os.makedirs(path)
                    except PermissionError:  # pragma: no cover
                        print("ERROR: Could not create the directory. Please modify the path variable in the output "
                              "section of the configuration file")
                        sys.exit(1)
                    except FileNotFoundError:
                        print("ERROR: File not found. Please modify the path variable in the output section of the "
                              "configuration file")
                        sys.exit(1)
                    except Exception:  # pragma: no cover
                        print("ERROR: Please modify the path variable in the output section of the configuration file")
                        sys.exit(1)
                return os.path.abspath(path) + "/"
        except KeyError:
            return "./"

    def get_output_name(self):
        try:
            name = self._json_config["workspace"]["name"]
            assert isinstance(name, str), "name in output section must be a string"
            return name
        except KeyError:
            return "output"

    def get_create_log_file(self):
        try:
            create_log_file = self._json_config["workspace"]["create_log_file"]
            assert isinstance(create_log_file, bool), "create_log_file must be a boolean"
            return create_log_file
        except KeyError:
            return True

    def get_verbose_level(self):
        """Set the verbose level from the configuration file or set a default value. Users can specify just the first
        letter of the verbose level and Learner will correct the values.

        :return: returns and sets the verbose level, default in "INFO"
        """
        supported_verbose_levels = {"D": "DEBUG", "I": "INFO", "W": "WARNING", "E": "ERROR", "C": "CRITICAL"}
        try:
            verbose_level = self._json_config["workspace"]["verbose_level"]
            assert isinstance(verbose_level, str), "verbose level must be a string"
            verbose_level = verbose_level.upper()
            if verbose_level not in supported_verbose_levels.values():
                try:
                    return supported_verbose_levels[verbose_level[0]]
                except KeyError:
                    logging.exception("The verbose level %s is not understood", verbose_level)
                    sys.exit(1)
            return verbose_level
        except KeyError:
            return "INFO"

    def setup_logging(self, logging_path='./schema/logging.json'):
        """Setup the logging by reading the default configuration file and updating the values using the values passed
        to the configuration file

        :param logging_path: the path to the default logging configuration file
        :return: None
        """
        log_directory = self._create_log_directory()
        logging_config = load_json(logging_path)
        self._update_logger_config(logging_config, log_directory)
        logging.captureWarnings(True)
        logging.config.dictConfig(logging_config)

    def _create_log_directory(self):
        """Create a directory to save the log files there

        :return: the path to the log directory
        """
        log_directory = self.output_path + "log/"
        if self.create_log_file:
            if not os.path.exists(log_directory):  # pragma: no cover
                os.makedirs(log_directory)
        return log_directory

    def _update_logger_config(self, logger_config, log_directory):
        """Update the logger configuration using the verbose_level passed through the configuration file as well as
        create_log_file flag

        :param logger_config: a dictionary containing the logging configuration
        :param log_directory: the path to the log directory
        :return: None
        """
        logger_config["loggers"][""]["level"] = self.verbose_level
        if self.create_log_file:  # pragma: no cover
            logger_config["handlers"]["file"]["filename"] = log_directory + self.output_name + str(self.sep_timetag) + ".txt"
        else:
            del logger_config["handlers"]["file"]
            logger_config["loggers"][""]["handlers"] = ["console"]


def load_json(source):
    """Load the json file.

    :param source: the path to a json file
    :return: the loaded json object
    """
    try:
        # first try to use the builtin library to load the file assuming it doesn't contain any comments. If it fails,
        # we'll use commentjson to load the file
        with open(source, "r") as stream:
            json_config = json.load(stream)
    except Exception:
        with open(source, "r") as stream:
            json_config = commentjson.load(stream)
    return json_config



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for validating a data set against the requirements that are either defined in a separate
metadata file or provided by user in the configuration file"""

import sys
import warnings
import logging
import valideer as v
import numpy as np
import pandas as pd


class DataValidator:
    def __init__(self, conf, data: pd, data_type="train"):
        """Accept a dataset and a conf object to validate the data set against the information provided in the conf
        object

        :param conf: a conf object (an instance of the Configuration class in configuration module)
        :param data: a pandas dataframe
        :param data_type: the data type, it can be "train", "validation" or "test"
        """
        self._conf = conf
        self._data = data
        self._data_type = data_type

    @property
    def conf(self):
        return self._conf

    @property
    def data(self):
        return self._data

    @property
    def data_type(self):
        return self._data_type

    def validate_against_meta_data(self, meta_data):
        """Validate the data against a meta_data file

        :param meta_data: the meta_data file
        :return: None
        """
        logging.info("Validating the data against the meta_data file...")

        # loop through columns and create new dict with just column and schema needed for validation
        schema_dict = dict()
        for col in meta_data['column']:
            for key, value in meta_data['column'][col].items():
                if key == 'schema':
                    schema_dict[col] = [value]
                    
        val = v.parse(schema_dict)
        for col in self._data.columns:
            if not val.is_valid(self._data[[col]].dropna().to_dict('list')):
                raise Exception("Data in column {0} is Not Valid, type expected: {1}".format(col, schema_dict[col]))

        logging.info("Finished validating against meta_data...")

    @staticmethod
    def check_nulls_in_col(data, col):
        """Check if the target column has any missing values.

        :param data: a pandas dataframe
        :param col: the column to check for missing values. This is usually the target column
        :return: None
        """
        logging.info("Checking if the target contains missing values...")
        try:
            num_nulls = data[col].isnull().sum()
            if num_nulls:
                data.dropna(subset=[col], inplace=True)
                warnings.warn("The target column contains {0} rows with missing values. Those rows will be dropped "
                              "from the dataset".format(num_nulls), Warning)
                if data.empty:
                    logging.error("Looks like all values in the target column are missing, please check your data."
                                  " Exiting...")
                    sys.exit(1)

        except KeyError:
            logging.info("target_col is not in the data or not loaded. Skipping check_nulls_in_target...")

    def check_dtype_of_num_cols(self):
        """Ensure the columns passed as numerical columns are actually numeric. Learner only issues a warning if it
        finds out some columns are not numeric.

        :return: None
        """
        if self._conf.process.to_numeric_cols and self._conf.process.to_numeric_activate:
            logging.info("Checking data types of numerical columns...")
            # get columns with numeric datatypes
            numeric_columns_types = [np.issubdtype(self.data[col].dtype, np.number)
                                     for col in self._conf.process.to_numeric_cols]
            # first check to see any num_cols is defined, if not just return the data
            # if num_cols is defined, make sure all passed columns are of type number (int and float).
            if not all(numeric_columns_types):
                warnings.warn("Column(s) passed {0} can't be converted to numeric data type. This may cause some "
                              "errors. Check the data".format([col_name for col_name in self._conf.process.to_numeric_cols
                                                               if np.issubdtype(self.data[col_name].dtype, np.number)
                                                               is False]), UserWarning)

    def check_drop_from_train_includes_id_cols(self):
        """Check if drop_from_train columns include id_columns. If not issue a warning. This is important because the
        id columns should not usually be included in training.

        :return: None
        """
        logging.info("Checking if drop_from_train includes id_cols...")
        if self._conf.column.id_cols:
            diff = set(self._conf.column.id_cols) - set(self._conf.column.drop_from_train)
            if diff:
                warnings.warn("The columns {0} exist in id_cols but not in drop_from_train. This may affect the "
                              "trained model.".format(diff), UserWarning)

    def validate_data(self):
        """The main function that runs all the instance methods if the validation flag is set to true

        :return: None
        """
        logging.info("Validating the data")

        if self._conf.data.meta_data_file:
            self.validate_against_meta_data(self._conf.data.meta_data)

        # we care about nulls in target only when data_type is train or validation
        if self.data_type != "test":
            self.check_nulls_in_col(self.data, self._conf.column.target_col)
        self.check_dtype_of_num_cols()
        self.check_drop_from_train_includes_id_cols()
        logging.info("Successfully validated the data")



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
import os
import warnings
import logging
from pathlib import Path
from importlib.util import spec_from_loader, module_from_spec
from importlib.machinery import SourceFileLoader

from learner.validator.input_validator import InputValidator, validate_subset_list
from learner.data_worker.data_loader import get_meta_data as data_worker_get_meta_data
from learner.configuration.workspace import WorkspaceConfiguration
from learner.configuration.sample import SampleConfiguration
from learner.configuration.supported_items import (SUPPORTED_DB_TYPES, SUPPORTED_SCORE_TYPES, SUPPORTED_INPUT_TYPES,
                                                   SUPPORTED_FILE_FORMATS)
from learner.data_worker.data_loader import get_value
from learner.setup.setup import load_json
from learner.utilities.templates import TRAIN_VALIDATION_SPLIT


class DataConfiguration:
    """Parse all input variables in the file section including the location of the train/test files, nrows, etc."""

    def __init__(self, json_config, workspace: WorkspaceConfiguration, sample: SampleConfiguration, engine, sep_timetag):
        self._json_config = json_config
        self._workspace = workspace
        self._sample = sample
        self._engine = engine
        self.sep_timetag = sep_timetag

        self.train_format = self.get_train_format()
        self.validation_format = self.get_validation_format()
        self.test_format = self.get_test_format()

        self.sample_validation_filename = TRAIN_VALIDATION_SPLIT.format(output_name=self.output.name,
                                                                        dtype="validation",
                                                                        sep_timetag=self.sep_timetag,
                                                                        format=self.validation_format)

        self.sample_train_filename = TRAIN_VALIDATION_SPLIT.format(output_name=self.output.name,
                                                                   dtype="train",
                                                                   sep_timetag=self.sep_timetag,
                                                                   format=self.train_format)

        self.train_location = get_value(self._json_config, None, "data", "train_params", "location")
        self.validation_location = self.get_validation_location()
        self.test_location = get_value(self._json_config, None, "data", "test_params", "location")

        self.train_input_type = self.get_train_input_type()
        self.validation_input_type = self.get_validation_input_type()
        self.test_input_type = self.get_test_input_type()

        self.meta_data_file = get_value(self._json_config, None, "data", "meta_data_file")
        self.meta_data_schema_file = get_value(self._json_config,
                                               str(Path(__file__).resolve().parents[1] / "schema" / "meta_data.schema.json"),
                                               "data",
                                               "meta_data_schema_file")
        self.meta_data_schema = load_json(self.meta_data_schema_file)
        self.meta_data = self.get_meta_data()
        self.manifest = get_value(self._json_config, None, "data", "manifest")

        self.train_delimiter = self.get_train_delimiter()
        self.validation_delimiter = self.get_validation_delimiter()
        self.test_delimiter = self.get_test_delimiter()

        self.train_nrows = get_value(self._json_config, None, "data", "train_params", "nrows")
        self.validation_nrows = get_value(self._json_config, None, "data", "validation_params", "nrows")
        self.test_nrows = get_value(self._json_config, None, "data", "test_params", "nrows")

        self.train_header = self.get_train_header()
        self.validation_header = self.get_validation_header()
        self.test_header = self.get_test_header()

        self.train_sample_size = self.get_train_sample_size()
        self.validation_sample_size = self.get_validation_sample_size()
        self.test_sample_size = self.get_test_sample_size()

        self.train_sample_seed = get_value(self._json_config, 42, "data", "train_params", "sample_seed")
        self.validation_sample_seed = get_value(self._json_config, 42, "data", "validation_params", "sample_seed")
        self.test_sample_seed = get_value(self._json_config, 42, "data", "test_params", "sample_seed")

        self.test_chunksize = self.get_test_chunksize()

        self.train_query_activate = self.get_train_query_activate()
        self.validation_query_activate = self.get_validation_query_activate()
        self.test_query_activate = self.get_test_query_activate()

        self.train_db_type = self.get_train_db_type()
        self.validation_db_type = self.get_validation_db_type()
        self.test_db_type = self.get_test_db_type()

        self.train_query_file = self.get_train_query_file()
        self.validation_query_file = self.get_validation_query_file()
        self.test_query_file = self.get_test_query_file()

        self.train_query = self.get_train_query()
        self.validation_query = self.get_validation_query()
        self.test_query = self.get_test_query()

        self.train_query_save_to_file = get_value(self._json_config, False, "data", "train_params", "query_params", "save_to_file")
        self.validation_query_save_to_file = get_value(self._json_config, False, "data", "validation_params", "query_params", "save_to_file")
        self.test_query_save_to_file = get_value(self._json_config, False, "data", "test_params", "query_params", "save_to_file")

        self.validation_prediction_type = None
        self.validation_score_activate = self.get_validation_score_activate()
        self.validation_score_types = self.get_validation_score_types()
        self.validation_add_timetag = get_value(self._json_config, True, "data", "validation_params", "scoring_params", "add_timetag")
        self.validation_column_name = get_value(self._json_config, "prediction", "data", "validation_params", "scoring_params", "column_name")
        self.validation_join_cols = self.get_validation_join_cols()

        self.test_prediction_activate = self.get_test_prediction_activate()
        self.test_prediction_type = self.get_test_prediction_type()
        self.test_add_timetag = get_value(self._json_config, True, "data", "test_params", "prediction_params", "add_timetag")
        self.test_clean_up = get_value(self._json_config, True, "data", "test_params", "prediction_params", "clean_up")
        self.test_column_name = get_value(self._json_config, "prediction", "data", "test_params", "prediction_params", "column_name")

        self.train_batch_size = get_value(self._json_config, 64, "data", "train_params", "batch_size")
        self.validation_batch_size = get_value(self._json_config, 64, "data", "validation_params", "batch_size")
        self.test_batch_size = get_value(self._json_config, 64, "data", "test_params", "batch_size")

    @property
    def output(self):  # pragma: no cover
        return self._workspace

    @property
    def sample(self):
        return self._sample

    def get_validation_location(self):
        try:
            location = self._json_config["data"]["validation_params"]["location"]
            if self.sample.split_activate:
                logging.critical("train_test_split was activated and validation location was also provided. "
                                 "Your intention is not clear to Learner. Please either remove the validation location "
                                 "or set train_test_split to false. Exiting...")
                sys.exit(1)
            return location
        except KeyError:
            if self.sample.split_activate:
                return self.output.data_directory + self.sample_validation_filename
            return None

    def get_train_input_type(self):
        # we don't need to validate the location here because it happens somewhere else and the need for validation
        # depends on other parameters
        if self.train_location and os.path.exists(self.train_location):
            if os.path.isfile(self.train_location):
                return "file"
            elif os.path.isdir(self.train_location):
                return "folder"

    def get_validation_input_type(self):
        # we don't need to validate the location here because it happens somewhere else and the need for validation
        # depends on other parameters
        if self.validation_location and self.validation_location:
            if os.path.exists(self.validation_location):
                if os.path.isfile(self.validation_location):
                    return "file"
                elif os.path.isdir(self.validation_location):
                    return "folder"

    def get_test_input_type(self):
        # we don't need to validate the location here because it happens somewhere else and the need for validation
        # depends on other parameters
        if self.test_location and self.test_location:
            if os.path.exists(self.test_location):
                if os.path.isfile(self.test_location):
                    return "file"
                elif os.path.isdir(self.test_location):
                    return "folder"

    def get_meta_data(self):
        if self.meta_data_file:
            self.meta_data = data_worker_get_meta_data(self.meta_data_file)
            logging.info("---Validating the meta data---")
            InputValidator(self.meta_data, self.meta_data_schema)
            return self.meta_data
        return None

    def get_train_delimiter(self):
        try:
            delimiter = self._json_config["data"]["train_params"]["delimiter"]
            if delimiter == "" or delimiter == " ":
                warnings.warn("An empty string was passed for train delimiter, using the default delimiter ','",
                              UserWarning)
                return ","
            return delimiter
        except KeyError:
            return ","

    def get_validation_delimiter(self):
        try:
            delimiter = self._json_config["data"]["validation_params"]["delimiter"]
            if delimiter == "" or delimiter == " ":
                warnings.warn("An empty string was passed for validation delimiter, using the default delimiter ','",
                              UserWarning)
                return ","
            return delimiter
        except KeyError:
            logging.info(f"The delimiter for validation data was not defined. Setting it to train delimiter {self.train_delimiter}")
            return self.train_delimiter

    def get_test_delimiter(self):
        try:
            delimiter = self._json_config["data"]["test_params"]["delimiter"]
            if delimiter == "" or delimiter == " ":
                warnings.warn("An empty string was passed for test delimiter, using the default delimiter ','",
                              UserWarning)
                return ","
            return delimiter
        except KeyError:
            logging.info(f"The delimiter for test data was not defined. Setting it to train delimiter {self.train_delimiter}")
            return self.train_delimiter

    def get_train_header(self):
        if self.manifest:
            return None
        else:
            try:
                return self._json_config["data"]["train_params"]["header"]
            except KeyError:
                return 0

    def get_validation_header(self):
        if self.manifest:
            return None
        else:
            try:
                return self._json_config["data"]["validation_params"]["header"]
            except KeyError:
                logging.info(f"The header for validation data was not defined. Setting it to train header {self.train_header}")
                return self.train_header

    def get_test_header(self):
        if self.manifest:
            return None
        else:
            try:
                return self._json_config["data"]["test_params"]["header"]
            except KeyError:
                logging.info(f"The header for test data was not defined. Setting it to train header {self.train_header}")
                return self.train_header

    def get_train_format(self):
        try:
            format_ = self._json_config["data"]["train_params"]["format"]
            validate_subset_list(parent_list=SUPPORTED_FILE_FORMATS, parent_name="supported file formats",
                                 subset_list=[format_], subset_name="defined file format")
            return format_
        except KeyError:
            return "csv"

    def get_validation_format(self):
        try:
            format_ = self._json_config["data"]["validation_params"]["format"]
            validate_subset_list(parent_list=SUPPORTED_FILE_FORMATS, parent_name="supported file formats",
                                 subset_list=[format_], subset_name="defined file format")
            return format_
        except KeyError:
            return "csv"

    def get_test_format(self):
        try:
            format_ = self._json_config["data"]["test_params"]["format"]
            validate_subset_list(parent_list=SUPPORTED_FILE_FORMATS, parent_name="supported file formats",
                                 subset_list=[format_], subset_name="defined file format")
            return format_
        except KeyError:
            return "csv"

    def get_train_sample_size(self):
        try:
            sample_size = self._json_config["data"]["train_params"]["sample_size"]
            if self.train_nrows is not None:
                warnings.warn("Both nrows and sample_size are defined for train data. Skipping nrows and using sample_size...",
                              UserWarning)
            return sample_size
        except KeyError:
            return None

    def get_validation_sample_size(self):
        try:
            sample_size = self._json_config["data"]["validation_params"]["sample_size"]
            if self.validation_nrows is not None:
                warnings.warn("Both nrows and sample_size are defined for validation data. Skipping nrows and using sample_size...",
                              UserWarning)
            return sample_size
        except KeyError:
            return None

    def get_test_sample_size(self):
        try:
            sample_size = self._json_config["data"]["test_params"]["sample_size"]
            if self.test_nrows is not None:
                warnings.warn("Both nrows and sample_size are defined for test data. Skipping nrows and using sample_size...",
                              UserWarning)
            return sample_size
        except KeyError:
            return None

    def get_test_chunksize(self):
        try:
            min_chunksize = 20000
            chunksize = self._json_config["data"]["test_params"]["chunksize"]
            if chunksize < min_chunksize:
                warnings.warn(f"chunksize was set to {chunksize} which is smaller than {min_chunksize}. Using chunksize = {min_chunksize} "
                              "for making predictions...", UserWarning)
                return min_chunksize
            return chunksize
        except KeyError:
            return 200000

    def get_train_query_activate(self):
        try:
            activate = self._json_config["data"]["train_params"]["query_params"]["activate"]
            if activate and "query" not in SUPPORTED_INPUT_TYPES[self._engine]:
                logging.critical(f"train query_params was activated for {self._engine} but this engine does not support "
                                 f"queries. Please update your configuration file and try again. Exiting...")
                sys.exit(1)
            if activate and self.train_location:
                logging.critical("train location was provided and train query_params was also activated. Learner "
                                 "does not know how to proceed in this situation. Please provide one of these "
                                 "options. Exiting...")
                sys.exit(1)
            return activate
        except KeyError:
            return False

    def get_validation_query_activate(self):
        try:
            activate = self._json_config["data"]["validation_params"]["query_params"]["activate"]
            if activate and "query" not in SUPPORTED_INPUT_TYPES[self._engine]:
                logging.critical(f"validation query_params was activated for {self._engine} but this engine does not support "
                                 f"queries. Please update your configuration file and try again. Exiting...")
                sys.exit(1)
            if activate and self.validation_location:
                logging.critical("validation location was provided and validation query_params was also activated. "
                                 "Learner does not know how to proceed in this situation. Please provide one of these "
                                 "options. Exiting...")
                sys.exit(1)
            return activate
        except KeyError:
            return False

    def get_test_query_activate(self):
        try:
            activate = self._json_config["data"]["test_params"]["query_params"]["activate"]
            if activate and "query" not in SUPPORTED_INPUT_TYPES[self._engine]:
                logging.critical(f"test query_params was activated for {self._engine} but this engine does not support "
                                 f"queries. Please update your configuration file and try again. Exiting...")
                sys.exit(1)
            if activate and self.test_location:
                logging.critical("""
                test location was provided and test query_params was also activated. Learner does not know how to 
                proceed in this situation. Please provide one of these options. Exiting...
                """)
                sys.exit(1)
            return activate
        except KeyError:
            return False

    def get_train_db_type(self):
        if self.train_query_activate:
            try:
                db_type = self._json_config["data"]["train_params"]["query_params"]["db_type"].lower()
                validate_subset_list(parent_list=SUPPORTED_DB_TYPES, parent_name="supported db types",
                                     subset_list=[db_type], subset_name="defined db type")
                return db_type
            except KeyError:
                return "presto"

    def get_validation_db_type(self):
        if self.validation_query_activate:
            try:
                db_type = self._json_config["data"]["validation_params"]["query_params"]["db_type"].lower()
                validate_subset_list(parent_list=SUPPORTED_DB_TYPES, parent_name="supported db types",
                                     subset_list=[db_type], subset_name="defined db type")
                return db_type
            except KeyError:
                return "presto"

    def get_test_db_type(self):
        if self.test_query_activate:
            try:
                db_type = self._json_config["data"]["test_params"]["query_params"]["db_type"].lower()
                validate_subset_list(parent_list=SUPPORTED_DB_TYPES, parent_name="supported db types",
                                     subset_list=[db_type], subset_name="defined db type")
                return db_type
            except KeyError:
                return "presto"

    def get_train_query_file(self):
        if self.train_query_activate:
            try:
                query_file = self._json_config["data"]["train_params"]["query_params"]["query_file"]
                if not os.path.isfile(query_file):
                    logging.critical("query_file is defined in the train_params section but it does not point to "
                                     "a valid file. Please update the configuration file and try again. Exiting...")
                    sys.exit(1)
                return query_file
            except KeyError:
                return None

    def get_validation_query_file(self):
        if self.validation_query_activate:
            try:
                query_file = self._json_config["data"]["validation_params"]["query_params"]["query_file"]
                if not os.path.isfile(query_file):
                    logging.critical("query_file is defined in the validation_params section but it does not point to "
                                     "a valid file. Please update the configuration file and try again. Exiting...")
                    sys.exit(1)
                return query_file
            except KeyError:
                return None

    def get_test_query_file(self):
        if self.test_query_activate:
            try:
                query_file = self._json_config["data"]["test_params"]["query_params"]["query_file"]
                if not os.path.isfile(query_file):
                    logging.critical("query_file is defined in the test_params section but it does not point to "
                                     "a valid file. Please update the configuration file and try again. Exiting...")
                    sys.exit(1)
                return query_file
            except KeyError:
                return None

    def get_train_query(self):
        if self.train_query_activate:
            try:
                return self._json_config["data"]["train_params"]["query_params"]["query"]
            except KeyError:
                # get the query from the query_file if it was provided
                if self.train_query_file:
                    spec = spec_from_loader("query_file", SourceFileLoader("query_file", self.train_query_file))
                    query_file = module_from_spec(spec)
                    spec.loader.exec_module(query_file)
                    try:
                        return query_file.train_query.strip()
                    except Exception as e:
                        logging.critical(f"Learner is not able to load the train_query from the query file. "
                                         f"The error is {e}. Exiting...")
                        sys.exit(1)
                logging.critical("""
                Learner was instructed to get train data using a query but no query or query_file was provided.
                Please update the configuration file and retry. Exiting...
                """)
                sys.exit(1)

    def get_validation_query(self):
        if self.validation_query_activate:
            try:
                return self._json_config["data"]["validation_params"]["query_params"]["query"]
            except KeyError:
                # get the query from the query_file if it was provided
                if self.validation_query_file:
                    spec = spec_from_loader("query_file", SourceFileLoader("query_file", self.validation_query_file))
                    query_file = module_from_spec(spec)
                    spec.loader.exec_module(query_file)
                    try:
                        return query_file.validation_query.strip()
                    except Exception as e:
                        logging.critical(f"Learner is not able to load the validation_query from the query file. "
                                         f"The error is {e}. Exiting...")
                        sys.exit(1)
                logging.critical("""
                Learner was instructed to get validation data using a query but no query or query_file was provided.
                Please update the configuration file and retry. Exiting...
                """)
                sys.exit(1)

    def get_test_query(self):
        if self.test_query_activate:
            try:
                return self._json_config["data"]["test_params"]["query_params"]["query"]
            except KeyError:
                # get the query from the query_file if it was provided
                if self.test_query_file:
                    spec = spec_from_loader("query_file", SourceFileLoader("query_file", self.test_query_file))
                    query_file = module_from_spec(spec)
                    spec.loader.exec_module(query_file)
                    try:
                        return query_file.test_query.strip()
                    except Exception as e:
                        logging.critical(f"Learner is not able to load the test_query from the query file. "
                                         f"The error is {e}. Exiting...")
                        sys.exit(1)
                logging.critical("""
                Learner was instructed to get test data using a query but no query or query_file was provided.              
                Please update the configuration file and retry. Exiting...
                """)
                sys.exit(1)

    def get_validation_score_activate(self):
        try:
            score = self._json_config["data"]["validation_params"]["scoring_params"]["activate"]
            if score and not self.sample.split_activate and not self.validation_query_activate and not self.validation_location:
                logging.error("It was requested to score the validation data but no query or location is provided "
                              "for the validation data, and train data is not being split. Please update the "
                              "configuration file and try again. Exiting...")
                sys.exit(1)
            return score
        except KeyError:
            return False

    def get_validation_score_types(self):
        score_types_dict = {}
        try:
            score_types_list = self._json_config["data"]["validation_params"]["scoring_params"]["types"]
            if score_types_list:
                if self.validation_score_activate is False:
                    warnings.warn("score types is not empty and scoring_params is not activated.", UserWarning)
                    return {}
                for score_type in score_types_list:
                    if isinstance(score_type, dict):
                        try:
                            score_types_dict[score_type["type"]] = score_type["params"]
                        except KeyError:
                            logging.exception("When passing a dictionary as an item to score_types, both type and "
                                              "params key must be present. The value of params can be a empty "
                                              "dictionary. Scoring type can also be a single string if you there "
                                              "is no extra parameters for the score type Exiting...")
                            sys.exit(1)
                    elif isinstance(score_type, str):
                        score_types_dict[score_type] = {}
                    else:
                        logging.error("Elements in score types must be either a string or a dictionary. "
                                      "Exiting...")
                        sys.exit(1)
        except KeyError:
            if self.validation_score_activate:
                warnings.warn("scoring_params was activated but no score types was defined.", UserWarning)
                self.validation_score_activate = False
            return score_types_dict

        self.validate_score_types(score_types_dict)
        self.update_validation_prediction_type(score_types_dict)
        return score_types_dict

    def get_validation_join_cols(self):
        try:
            join_cols = self._json_config["data"]["validation_params"]["scoring_params"]["join_cols"]
        except KeyError:
            join_cols = []
        return join_cols

    def validate_score_types(self, score_types_dict):
        if self.validation_score_activate:
            for score_type, params in score_types_dict.items():
                if score_type not in SUPPORTED_SCORE_TYPES[self._engine]:
                    if "file" in params:
                        self._get_score_function(score_type, params)
                    else:
                        logging.error("Invalid types in 'scoring_params', supported score types for %s engine "
                                      "are %s. Exiting...", self._engine, sorted(SUPPORTED_SCORE_TYPES[self._engine]))
                        sys.exit(1)

    def _get_score_function(self, score_type, params):
        spec = spec_from_loader("score_file", SourceFileLoader("score_file", params["file"]))
        score_file = module_from_spec(spec)
        spec.loader.exec_module(score_file)
        try:
            params["function"] = getattr(score_file, score_type)
        except Exception as e:
            logging.critical(f"Learner is not able to load the score function from the score file. "
                             f"The error is {e}. Exiting...")
            sys.exit(1)

    def update_validation_prediction_type(self, score_types_dict):
        required_pred_type = set()
        if self.validation_score_activate:
            for score_type, params in score_types_dict.items():
                if "file" not in params:
                    required_pred_type.add(SUPPORTED_SCORE_TYPES[self._engine][score_type])
                    if len(required_pred_type) > 1:
                        self.validation_prediction_type = "all"
                    else:
                        self.validation_prediction_type = list(required_pred_type)[0]

    def get_test_prediction_activate(self):
        try:
            activate = self._json_config["data"]["test_params"]["prediction_params"]["activate"]
            if activate:
                if self.test_location:
                    if ("folder" in SUPPORTED_INPUT_TYPES[self._engine] and not os.path.exists(self.test_location)) or \
                       ("folder" not in SUPPORTED_INPUT_TYPES[self._engine] and not os.path.isfile(self.test_location)):
                        logging.error("prediction_params was activated but the test_location does not point "
                                      "to a valid location. Exiting...")
                        sys.exit(1)
                elif not self.test_query_activate and "query" in SUPPORTED_INPUT_TYPES[self._engine]:
                    logging.error("prediction_params was activated . However, no location or query was Provided. "
                                  "Exiting...")
                    sys.exit(1)
            return activate
        except KeyError:
            return False

    def get_test_prediction_type(self):
        supported_prediction_types = ['proba', 'class', 'all']
        try:
            prediction_type = self._json_config["data"]["test_params"]["prediction_params"]["type"]
            # guess the prediction_type based on the first few characters
            if prediction_type.lower().startswith('c'):
                prediction_type = "class"
            if prediction_type.lower().startswith('p'):
                prediction_type = "proba"
            if prediction_type.lower().startswith('a'):
                prediction_type = "all"

            if not any(prediction_type.lower().startswith(p_type) for p_type in supported_prediction_types):
                logging.error("Prediction type provided is invalid. Please choose from the following options: %s",
                              supported_prediction_types)
                sys.exit(1)
            return prediction_type
        except KeyError:
            return "proba"




# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module to compute the prediction scores using various metrics."""

import sys
import logging
import warnings

import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score, f1_score, average_precision_score, precision_score, recall_score
from sklearn.metrics import roc_auc_score, brier_score_loss, classification_report, confusion_matrix, fbeta_score
from sklearn.metrics import hamming_loss, jaccard_similarity_score, log_loss, matthews_corrcoef, precision_recall_curve
from sklearn.metrics import precision_recall_fscore_support, zero_one_loss
from sklearn.metrics import explained_variance_score, mean_absolute_error, mean_squared_error, mean_squared_log_error
from sklearn.metrics import median_absolute_error, r2_score

from learner.model_manager.scorers import root_mean_squared_error
from learner.data_worker.data_loader import get_data
from learner.configuration.supported_items import SUPPORTED_SCORE_TYPES
from learner.configuration.configuration import Configuration
from learner.data_worker.data_processor import delete_keys_from_dict


class BaseScorer:
    """Score the prediction results against target col in test data"""

    def __init__(self, conf, models_dict, processor=None):
        """Initialize a BaseScorer instance using a conf object and models_dict.

        :param conf: configuration object
        :param models_dict: learner models dictionary
        """
        self._conf = conf
        self._models_dict = models_dict
        self._processor = processor
        self._y_true = self.get_y_true()

    @property
    def conf(self):
        return self._conf

    @property
    def models_dict(self):
        return self._models_dict

    @property
    def y_true(self):
        return self._y_true

    def get_y_true(self):
        """Load the target col from the test data.

        :return y_true: target col along with the join columns
        """
        logging.info("Loading target col from the validation data...")
        y_true = get_data(self._conf.data.validation_location,
                          manifest_file=self._conf.data.manifest,
                          format=self._conf.data.validation_format,
                          sep=self._conf.data.validation_delimiter,
                          nrows=self._conf.data.validation_nrows,
                          header=self._conf.data.validation_header,
                          usecols=[self._conf.column.target_col] + self.conf.data.validation_join_cols)

        # if we need to score the log transformed values, we do the transformation here
        if self.conf.process.log_transform_target_activate and not self.conf.process.log_transform_target_score_actuals:
            from learner.data_worker.data_processor import log_transform
            y_true = log_transform(data=y_true, params=self._conf.process.log_transform_target_params, cols=[self._conf.column.target_col])
        # in case target column was label encoded, use the trained label encoder in the processor object to
        # label encode y_true
        if self._processor and self._conf.process.label_encode_target:
            self._processor.handle_label_encoding(data=y_true, cols=[self._conf.column.target_col])
        return y_true

    def scoring_handler(self, score_type, y_true, y_pred, params):
        """Construct the call to appropriate methods to score the predictions.

        :param score_type: the metric for computing the score
        :param y_true: a dataframe containing the true values
        :param y_pred: the estimated target
        :param params: additional parameters to pass to the scoring function
        :return: the computed score
        """
        # this method accepts numpy arrays and pandas dataframe but we convert np arrays or pd series to pd dataframe
        if isinstance(y_true, np.ndarray) or isinstance(y_true, pd.Series):
            y_true = pd.DataFrame(data=y_true)
        if isinstance(y_pred, np.ndarray) or isinstance(y_pred, pd.Series):
            y_pred = pd.DataFrame(data=y_pred)

        # this is number of items before removing nulls
        before_cnt = y_true.shape[0]

        # We remove any missing values. y_pred should not have any but we check it just to be sure
        with warnings.catch_warnings():
            # this operations raises a warning and we want to suppress it
            warnings.simplefilter("ignore")
            y_true = y_true[y_true.notna().any(1) & y_pred.notna().any(1)]
            y_pred = y_pred[y_true.notna().any(1) & y_pred.notna().any(1)]

        # this is number of items after removing nulls
        after_cnt = y_true.shape[0]
        removed_cnt = before_cnt - after_cnt
        if removed_cnt > 0:
            warnings.warn(f"There were {removed_cnt} missing values in the data when scoring.", UserWarning)
        # handle custom scores first
        if params and "file" in params:
            params_copy = params.copy()
            delete_keys_from_dict(params_copy, ["file", "function"])
            return params["function"](y_true, y_pred, **params_copy)
        try:
            return getattr(sys.modules[__name__], score_type+"_score")(y_true, y_pred, **params)
        except AttributeError:
            return getattr(sys.modules[__name__], score_type)(y_true, y_pred, **params)


class ClassifierScorer(BaseScorer):
    def __init__(self, conf, models_dict, processor=None):
        """Initialize a ClassifierScorer instance using a conf object and models_dict.

        :param conf: configuration object
        :param models_dict: learner models dictionary
        """
        super().__init__(conf, models_dict, processor)

        # pandas dataframe to hold the prediction for class and probabilities
        self._pred_class = None
        self._pred_proba = None
        # these will hold the column names for class and proba
        self._proba_col_names = None
        self._class_col_names = None

    @property
    def pred_class(self):
        return self._pred_class

    @property
    def pred_proba(self):
        return self._pred_proba

    def load_pred_class(self, mdl, filename):
        """Load prediction class from the prediction file.

        :param mdl: a models_dict item
        :param filename: the full path to the prediction file
        :return: None
        """
        logging.info("Loading prediction classes from the output file...")
        if self._conf.data.validation_prediction_type == "all":
            self._class_col_names = list(filter(lambda x: x.startswith('_class_'), mdl["validation_pred_cols"]))
        else:
            self._class_col_names = list(mdl["validation_pred_cols"])

        self._pred_class = get_data(filename,
                                    manifest_file=None,
                                    usecols=self._class_col_names + self.conf.data.validation_join_cols)

    def load_pred_proba(self, mdl, filename):
        """Load prediction probability from the prediction file.

        :param mdl: a models_dict item
        :param filename: the full path to the prediction file
        :return: None
        """
        logging.info("Loading prediction probabilities from the output file...")
        if self._conf.data.validation_prediction_type == "all":
            self._proba_col_names = list(filter(lambda x: x.startswith('_proba_'), mdl["validation_pred_cols"]))
        else:
            self._proba_col_names = list(mdl["validation_pred_cols"])

        self._pred_proba = get_data(filename,
                                    manifest_file=None,
                                    usecols=self._proba_col_names + self.conf.data.validation_join_cols)

    def update_true_and_pred(self, pred, col_name):
        """In situations where the order of data in the prediction dataset and test data are different, we need to
        merge these two data on some columns to ensure the score values are valid. This method handles that. If there
        are some id columns to merge these two datasets, it merges them , and updates the prediction and true values.
        It accepts the prediction dataframe and a list of column name; if score_id_cols is empty, simply return the
        prediction dataframe and y_true with only target column for scoring. If score_id_cols is not empty,
        merge the prediction dataframe and y_true dataframe on score_id_cols, then return the prediction cols and
        target column from this merged dataset to be the prediction and target values. This handles the situations in
        which the order of prediction and target datasets are different. This is typically the case when we have
        segmentation.

        :param pred: the prediction dataframe. This can have multiple columns.
        :param col_name: the name of the columns that contain prediction values/
        :return: the updated prediction dataset (updated order) and the target column (possibly with updated order)
        """
        if not self.conf.data.validation_join_cols:
            return pred[col_name], self._y_true[self._conf.column.target_col]

        df = pred.merge(self._y_true, on=self.conf.data.validation_join_cols, how="inner")

        y_true = df[self._conf.column.target_col]
        pred = df[col_name]

        return pred, y_true

    def score(self):
        """The main function to compute the scores for the predictions generated by all models.

        :return: None
        """
        for tag, mdls in self._models_dict.items():
            mdl = mdls.get(0, mdls)
            logging.info("Score the predictions for model %s", tag)
            filename = mdl["path"] + self._conf.workspace.name + \
                "_validation_" + str(tag) + str(self._conf.sep_timetag) + ".csv"

            self._pred_class = None
            self._pred_proba = None

            model_score_dict = self.get_model_score_dict(mdl, filename)
            for score_type, score in model_score_dict.items():
                logging.info("%s model's %s score = %s", tag, score_type, score)

    def get_model_score_dict(self, mdl, filename):
        """Make the call to the appropriate functions to calculate all requested scores for a model.

        :param mdl: a models_dict item
        :param filename: the full path to the prediction file
        :return: mode_score_dict that contains the score types and their corresponding values
        """
        model_score_dict = {}
        for score_type, params in self._conf.data.validation_score_types.items():

            if SUPPORTED_SCORE_TYPES[self._conf.engine][score_type] == "class":
                if self._pred_class is None:
                    self.load_pred_class(mdl, filename)
                    self._pred_class, y_true = self.update_true_and_pred(self._pred_class, self._class_col_names)

                model_score_dict[score_type] = self.scoring_handler(score_type, y_true, self._pred_class, params)

            if SUPPORTED_SCORE_TYPES[self._conf.engine][score_type] == "proba":
                if self._pred_proba is None:
                    self.load_pred_proba(mdl, filename)
                    self._pred_proba, y_true = self.update_true_and_pred(self._pred_proba, self._proba_col_names)

                model_score_dict[score_type] = self.scoring_handler(score_type, y_true, self._pred_proba, params)

        return model_score_dict


class RegressorScorer(BaseScorer):
    def __init__(self, conf: Configuration, models_dict, processor=None):
        """Initialize a RegressorScorer instance using a conf object and models_dict.

        :param conf: configuration object
        :param models_dict: learner models dictionary
        """
        super().__init__(conf, models_dict, processor)
        self._pred = None

    @property
    def pred(self):
        return self._pred

    def load_pred(self, filename):
        """Load prediction class from the prediction file

        :param filename: the full path to the prediction file
        :return: None
        """
        logging.info("Loading prediction column from the output file...")

        self._pred = get_data(filename,
                              manifest_file=None,
                              usecols=[self._conf.data.validation_column_name] + self.conf.data.validation_join_cols)

    def update_true_and_pred(self):
        """In situations where the order of data in the prediction dataset and test data are different, we need to
        merge these two data on some columns to ensure the score values are valid. This method handles that. If there
        are some id columns to merge these two datasets, it merges them , and updates the prediction and true values.

        :return: None
        """
        if not self.conf.data.validation_join_cols:
            return
        df = self._pred.merge(self._y_true, on=self.conf.data.validation_join_cols, how="inner")

        self._y_true = df[self._conf.column.target_col]
        self._pred = df[[self._conf.data.validation_column_name]]

    def score(self):
        """The main function to compute the scores for the predictions generated by all models

        :return: None
        """
        for tag, mdls in self._models_dict.items():
            mdl = mdls.get(0, mdls)
            logging.info("Score the predictions for model %s", tag)
            filename = mdl["path"] + self._conf.workspace.name + \
                "_validation_" + str(tag) + str(self._conf.sep_timetag) + ".csv"

            self._pred = None

            model_score_dict = self.get_model_score_dict(filename)
            for score_type, score in model_score_dict.items():
                logging.info("%s model's %s score = %s", tag, score_type, score)

    def get_model_score_dict(self, filename):
        """Make the call to the appropriate functions to calculate all requested scores for a model.

        :param filename: the full path to the prediction file
        :return: mode_score_dict that contains the score types and their corresponding values
        """
        model_score_dict = {}
        for score_type, params in self._conf.data.validation_score_types.items():

            if self._pred is None:
                self.load_pred(filename)
                self.update_true_and_pred()
            model_score_dict[score_type] = self.scoring_handler(score_type, self._y_true, self._pred, params)

        return model_score_dict


class DeepClassifierScorer(BaseScorer):
    """The scorer class for the image_classifier engine. In image_classifier engine, scoring the predictions happens as
    we iterate through the validation data. This class is provided with y_true and the probabilities. It then uses
    those values and the user input to compute the scores. This class also updates the "metrics" dictionary as it
     computes he scores so it could later be used to log the numbers."""
    def __init__(self, conf: Configuration, models_dict, y_true, pred_proba, metrics=dict):
        """Initialize a ImageClassifierScorer using the input parameters.

        :param conf: a conf object
        :param models_dict: the models dict. This is set to None and is only used to be passed to the parent class.
        :param y_true: a numpy array that contains the y true.
        :param pred_proba: a numpy array that contains the probabilities of the predictions for each class.
        :param metrics: a dictionary that contains the metrics for each epoch
        """
        self._y_true = y_true
        super().__init__(conf, models_dict)
        self._pred_class = self.get_pred_class(pred_proba)
        self._pred_proba = self.update_pred_proba(pred_proba)
        self.metrics = metrics

    def get_y_true(self):
        """Convert y_true to a pandas series.

        :return: a pandas series created from the y_true numpy array.
        """
        y_true = pd.Series(data=self._y_true, name="y_true")

        # in case target column was label encoded, use the trained label encoder in the processor object to
        # label encode y_true
        if self._processor and self._conf.process.label_encode_target:
            self._processor.handle_label_encoding(data=y_true, cols=[self._conf.column.target_col])

        return y_true

    def update_pred_proba(self, pred_proba):
        """Create a pandas dataframe from the pred_proba numpy array. If we are doing a binary classification (two
        columns), take a second column (the probability of class 1)

        :param pred_proba: a numpy array containing the probability of predictions.
        :return: a pandas dataframe created using the pred_proba numpy array
        """
        pred_proba = pd.DataFrame(pred_proba)
        if pred_proba.shape[1] == 2:
            return pred_proba.iloc[:, 1]
        return pred_proba

    def get_pred_class(self, pred_proba):
        """Use the probabilities to get the predicted class. The predicted class is the index that has the highest
        probability (argmax)

        :param pred_proba: a numpy array containing the probability of predictions.
        :return: a numpy array contaning the predicted class/
        """
        return np.argmax(pred_proba, 1)

    def score(self):
        """The main function to compute the scores for the predictions generated by the model. Here, we first call
        get_model_score_dict to build a dictionary that contains the score types and their corresponding score. We then
        use that dictionary to update the metrics dictionary. That dictionary is then used by other classes (callbacks)
        to print the scores.

        :return: None
        """
        model_score_dict = self.get_model_score_dict()
        for score_type, score in model_score_dict.items():
            self.metrics[score_type] = score

    def get_model_score_dict(self):
        """Make the call to the appropriate functions to calculate all requested scores for a model.

        :return: mode_score_dict that contains the score types and their corresponding values
        """
        model_score_dict = {}
        for score_type, params in self._conf.data.validation_score_types.items():

            if SUPPORTED_SCORE_TYPES[self._conf.engine][score_type] == "class":
                model_score_dict[score_type] = self.scoring_handler(score_type, self._y_true, self._pred_class, params)

            if SUPPORTED_SCORE_TYPES[self._conf.engine][score_type] == "proba":
                model_score_dict[score_type] = self.scoring_handler(score_type, self._y_true, self._pred_proba, params)

        return model_score_dict


class DeepRegressorScorer(BaseScorer):
    """The scorer class for the deep_regressor engine. In deep_regressor engine, scoring the predictions happens as
    we iterate through the validation data. This class is provided with y_true and the predictions. It then uses
    those values and the user input to compute the scores. This class also updates the "metrics" dictionary as it
    computes the scores so it could later be used to log the numbers."""
    def __init__(self, conf: Configuration, models_dict, y_true, pred, metrics=dict):
        """Initialize a DeepRegressorScorer using the input parameters.

        :param conf: a conf object
        :param models_dict: the models dict. This is set to None and is only used to be passed to the parent class.
        :param y_true: a numpy array that contains the y true.
        :param pred: a numpy array that contains the probabilities of the predictions for each class.
        :param metrics: a dictionary that contains the metrics for each epoch
        """
        self._y_true = y_true
        super().__init__(conf, models_dict)
        self._pred = self.update_pred(pred)
        self.metrics = metrics

    def get_y_true(self):
        """Convert y_true to a pandas series. This method does a few other things as well. If we had log tranformed the
        target and we need to score the actuals, it does the exponential transformation.

        :return: a pandas series created from the y_true numpy array.
        """
        y_true = pd.DataFrame(data=self._y_true, columns=["y_true"])
        # if we need to score the actual values, we do the transformation here
        if self.conf.process.log_transform_target_activate and self.conf.process.log_transform_target_score_actuals:
            from learner.data_worker.data_processor import DataProcessor
            y_true = DataProcessor.exponential_transform(y_true,
                                                         params=[{"activate": True,
                                                                  "power": self.conf.process.log_transform_target_base,
                                                                  "col": "y_true",
                                                                  "name": "y_true",
                                                                  "shift": -self.conf.process.log_transform_target_shift}],
                                                         cols=["y_true"])

        return y_true

    def update_pred(self, pred):
        """Create a pandas dataframe from the pred numpy array. This method does a few other things as well. If we had log tranformed the
        target and we need to score the actuals, it does the exponential transformation.

        :param pred: a numpy array containing the probability of predictions.
        :return: a pandas dataframe created using the pred_proba numpy array
        """
        pred = pd.DataFrame(pred, columns=["pred"])
        if self.conf.process.log_transform_target_activate and self.conf.process.log_transform_target_score_actuals:
            from learner.data_worker.data_processor import DataProcessor
            pred = DataProcessor.exponential_transform(pred,
                                                       params=[{"activate": True,
                                                                "power": self.conf.process.log_transform_target_base,
                                                                "col": "pred",
                                                                "name": "pred",
                                                                "shift": -self.conf.process.log_transform_target_shift}],
                                                       cols=["pred"])

        return pred

    def score(self):
        """The main function to compute the scores for the predictions generated by the model. Here, we first call
        get_model_score_dict to build a dictionary that contains the score types and their corresponding score. We then
        use that dictionary to update the metrics dictionary. That dictionary is then used by other classes (callbacks)
        to print the scores.

        :return: None
        """
        model_score_dict = self.get_model_score_dict()
        for score_type, score in model_score_dict.items():
            self.metrics[score_type] = score

    def get_model_score_dict(self):
        """Make the call to the appropriate functions to calculate all requested scores for a model.

        :return: mode_score_dict that contains the score types and their corresponding values
        """
        model_score_dict = {}
        for score_type, params in self._conf.data.validation_score_types.items():

            model_score_dict[score_type] = self.scoring_handler(score_type, self._y_true, self._pred, params)

        return model_score_dict



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements classes and functions to build optimizer for training neural networks."""
import sys
from copy import deepcopy
import logging

from torch.optim import Adam, AdamW, SparseAdam, Adamax, ASGD, LBFGS, RMSprop, Rprop, SGD

from learner.data_worker.data_processor import delete_keys_from_dict


class OptimizerBuilder:
    """A class to build an optimizer based on the input parameters and model parameters.
    """
    def __init__(self, params, optimizer_dict):
        """Initialize an OptimizerBuilder object using a the model parameters (from the PyTorch model) and
        other input parameters that contains the optimizer type, lr, etc.

        :param params: the parameters of the model to be optimized. We may need to consider moving this to the build_optimizer method.
        :param optimizer_dict: a dictionary that contains all the parameters for instantiating the optimizer object
        """
        self._params = params
        self._optimizer_type = optimizer_dict["type"]
        self._optimizer_dict = deepcopy(optimizer_dict)

    @property
    def params(self):
        return self._params

    @property
    def optimizer_type(self):
        return self._optimizer_type

    @property
    def optimizer_dict(self):
        return self._optimizer_dict

    def build_optimizer(self):
        """The main method to get the optimizer object using functional programming. Here, we use the optimizer type to
        get the optimizer object.

        :return: the optimizer object with all the parameters set.
        """
        delete_keys_from_dict(self.optimizer_dict, ["type"])
        try:
            return getattr(sys.modules[__name__], self.optimizer_type)(self.params, **self.optimizer_dict)
        except Exception as e:
            logging.critical(f"Unable to build the optimizer. The error is: {e}")
            sys.exit(1)




# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
import os
import warnings
import logging

import torch

from learner.data_worker.data_loader import get_value
from learner.configuration.supported_items import SUPPORTED_MODELS, SUPPORTED_INPUT_TYPES, SUPPORTED_ENGINES_FOR_LEARNING_RATE_PARAMS


class ModelsConfiguration:
    """Parse all input variables related to the model section. The type of each model and the
    hyperprameters are defined here."""

    def __init__(self, json_config, sep_timetag, output, data, engine, segmenter):
        self._json_config = json_config
        self.sep_timetag = sep_timetag
        self._output = output
        self._data = data
        self._engine = engine
        self._segmenter = segmenter
        self.device = self.get_device()
        self.num_models = get_value(self._json_config, 1, "model", "num_models")
        self.train_models = self.get_train_models()

        self.cv_activate = self.get_cv_activate()
        self.cv_kfold_params = self.get_cv_kfold_params()
        self.cv_options = self.get_cv_options()

        self.gs_activate = self.get_gs_activate()
        self.gs_kfold_params = self.get_gs_kfold_params()
        self.gs_options = get_value(self._json_config, {}, "model", "grid_search_params", "options")

        self.lr_activate = self.get_lr_activate()
        self.lr_start_lr = get_value(self._json_config, 1e-7, "model", "learning_rate_params", "start_lr")
        self.lr_end_lr = get_value(self._json_config, 10, "model", "learning_rate_params", "end_lr")
        self.lr_num_it = get_value(self._json_config, 100, "model", "learning_rate_params", "num_it")
        self.lr_stop_div = get_value(self._json_config, True, "model", "learning_rate_params", "stop_div")

        self.save_models = self.get_save_models()
        self.models_dict = self.get_models_dict()
        # set nrows_score to None. As soon as the engine gets instantiated, the value will be set to 0
        # I don't set it here to avoid confusions
        self.nrows_score = None

        self.version_activate = get_value(self._json_config, False, "model", "version_params", "activate")
        self.version_column_name = get_value(self._json_config, "version", "model", "version_params", "column_name")
        self.version_version_name = self.get_version_version_name()

    def get_device(self):
        try:
            device = self._json_config["model"]["device"]
            if device == "cuda" and not torch.cuda.is_available():
                warnings.warn("device was set to cuda but it looks like no gpu is available. Setting the device to "
                              "cpu", UserWarning)
                return "cpu"
            return device
        except KeyError:
            if torch.cuda.is_available():
                return 'cuda'
            return 'cpu'

    def get_train_models(self):
        try:
            train_models = self._json_config["model"]["train_models"]
            if train_models:
                if self._data.train_location:
                    if ("folder" in SUPPORTED_INPUT_TYPES[self._engine] and not os.path.exists(self._data.train_location)) or\
                       ("folder" not in SUPPORTED_INPUT_TYPES[self._engine] and not os.path.isfile(self._data.train_location)):
                        logging.error("train_models was set to true, but the location value is not valid. "
                                      "Exiting...")
                        sys.exit(1)
                elif not self._data.train_query_activate and "query" in SUPPORTED_INPUT_TYPES[self._engine]:
                    logging.error("train_models was set to true. However, no location or query was provided. Exiting...")
                    sys.exit(1)
            return train_models
        except KeyError:
            if self._data.train_location:
                if ("folder" in SUPPORTED_INPUT_TYPES[self._engine] and not os.path.exists(self._data.train_location)) or\
                   ("folder" not in SUPPORTED_INPUT_TYPES[self._engine] and not os.path.isfile(self._data.train_location)):
                    logging.error("train_models defaults to true, but the location value is not valid.")
                    sys.exit(1)
            elif not self._data.train_query_activate and "query" in SUPPORTED_INPUT_TYPES[self._engine]:
                logging.error("train_models defaults to true, but neither location nor query were provided. Exiting...")
                sys.exit(1)
            return True

    def get_cv_activate(self):
        try:
            return self._json_config["model"]["cross_val_params"]["activate"]
        except KeyError:
            logging.info("Not using cross_validation...")
            return False

    def get_cv_kfold_params(self):
        try:
            return self._json_config["model"]["cross_val_params"]["kfold_params"]
        except KeyError:
            if self.cv_activate:
                warnings.warn("Cross validation was requested but k was not set, using default values: "
                              "{\"n_splits\": 3, \"shuffle\": false, \"random_state\": 42}", UserWarning)
            return {"n_splits": 3, "shuffle": False, "random_state": 42}

    def get_cv_options(self):
        # TODO: see if we should do something here depending on the input
        try:
            return self._json_config["model"]["cross_val_params"]["options"]
        except KeyError:
            return {}

    def get_gs_kfold_params(self):
        try:
            return self._json_config["model"]["grid_search_params"]["kfold_params"]
        except KeyError:
            if self.cv_activate:
                warnings.warn("Grid Search was requested but kfold_params was not set, using default values: "
                              "{\"n_splits\": 3, \"shuffle\": false, \"random_state\": 42}", UserWarning)
            return {"n_splits": 3, "shuffle": False, "random_state": 42}

    def get_lr_activate(self):
        try:
            activate = self._json_config["model"]["learning_rate_params"]["activate"]
            if activate and not self.train_models:
                logging.critical("learning_rate_params was activated but train_models was set to false. This feature "
                                 "is only valid when training the models. Please update the configuration file and "
                                 "try again. Exiting...")
                sys.exit(1)
            if activate and self._engine not in SUPPORTED_ENGINES_FOR_LEARNING_RATE_PARAMS:
                logging.critical("learning_rate_params was activated but it is not supported for the {engine} engine. "
                                 "The supported engines are {supported_engines}. Please update your configuration "
                                 "file. Exiting...".format(engine=self._engine,
                                                           supported_engines=SUPPORTED_ENGINES_FOR_LEARNING_RATE_PARAMS))
                sys.exit(1)
            return activate
        except KeyError:
            return False

    def get_save_models(self):
        try:
            return self._json_config["model"]["save_models"]
        except KeyError:
            logging.info("Not saving any models")
            return False

    def get_gs_activate(self):
        try:
            activate = self._json_config["model"]["grid_search_params"]["activate"]
            if self.cv_activate and activate:
                logging.error("Both do_cross_val and do_grid_search are set to true. Please set one of them to "
                              "false. Exiting..")
                sys.exit(1)
            return activate
        except KeyError:
            return False

    def get_models_dict(self):
        # if we are not training the models, we don't care about models_dict
        if not self.train_models:
            return {}

        def validate_mdl(mdl):
            try:
                if mdl["type"] not in SUPPORTED_MODELS[self._engine]:
                    logging.error("The model %s is not supported for %s engine. The supported models, "
                                  "in alphabetical order, are %s, Exiting...",
                                  mdl["type"], self._engine, sorted(SUPPORTED_MODELS[self._engine]))
                    sys.exit(1)
            except KeyError:
                if mdl["type"] not in SUPPORTED_MODELS["Classifier"]:
                    logging.error("The model %s is not supported. The supported models, "
                                  "in alphabetical order, are %s, Exiting...",
                                  mdl["type"], sorted(SUPPORTED_MODELS.values()))
                    sys.exit(1)

            mdl["path"] = directory_path

        try:
            models_dict = self._json_config["model"]["models_dict"]
            if self.num_models != len(models_dict):
                logging.error("Expected %i models but got %i. Exiting...", self.num_models, len(models_dict))
                sys.exit(1)
            for tag, mdls in models_dict.items():
                directory_path = self._output.path + self._output.name + "_" + tag + str(self.sep_timetag) + "/"
                # if a dictionary is passed, no need to loop
                if isinstance(mdls, dict):
                    validate_mdl(mdls)
                # loop through each item to validate
                elif isinstance(mdls, list):
                    if not self._segmenter.activate:
                        logging.critical(f"""
                            Segmenter was not activated but the parameters for {tag} was provided in a list format.
                            Please turn on segmener or provide a dictionary. Exiting...
                        """)
                        sys.exit(1)

                    for mdl in mdls:
                        validate_mdl(mdl)

                if not os.path.exists(directory_path) and self.train_models:  # pragma: no cover
                    os.makedirs(directory_path)
            return models_dict
        except KeyError:
            if self.train_models:
                logging.error("models_dict must be defined with train_models set to true...")
                sys.exit(1)
            return {}

    def get_version_version_name(self):
        if self.version_activate:
            try:
                version_name = self._json_config["model"]["version_params"]["version_name"]
                return version_name
            except KeyError:
                logging.critical('version_params in the model section was activated but no value for "version_name" '
                                 'was provided. Please update your configuration file and try again. Exiting...')
                sys.exit(1)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for computing similarity matrices. This module provides a wrapper around scipy distance package."""

import sys
import warnings
import logging
import scipy.spatial.distance as dist
import pandas as pd
import numpy as np


class Similarities:

    @staticmethod
    def compute_similarities(X, metric='jaccard', return_dataframe=True, cols=None, **kwargs):
        """A wrapper around scipy pdist function to compute similarity matrices.

        :param X: the input in binarized form
        :param metric: the metric to use (all metric supported by scipy can be used here)
        :param return_dataframe: a flag to define is the output should be returned in dataframe format
        :param cols: the columns to use if the return value is set to be a dataframe
        :param kwargs: kwargs to pass to pdist function
        :return: the similarity matrix

        Note: the scipy pdist computes the distances. 1 - dist is returned for similarity calculations
        """
        if X.shape[1] < 2:
            logging.error("The input matrix for similarity calculations must have more than 1 columns. Exiting...")
            sys.exit(1)

        # Note: transpose is needed to get the correct results
        matrix = 1 - dist.squareform(dist.pdist(X.T, metric, kwargs))

        if metric is 'jaccard' and np.array_equal(X, X.astype(bool)) is False:
            logging.error("Non-binary matrix is passed when calculating Jaccard distance."
                          " Make sure the matrix is binarized")
            sys.exit(1)

        if return_dataframe and cols is None:
            warnings.warn("No column names were passed to create the dataframe, "
                          "The default RangeIndex (0, 1, 2, …, n) will be used...", Warning)
            return pd.DataFrame(matrix)

        if return_dataframe and cols is not None:
            try:
                return pd.DataFrame(matrix, columns=cols)
            except ValueError:
                logging.exception("The length of the passed cols %s is %i, which doesn't match the expected length %i",
                                  cols, len(cols), matrix.shape[1])
                raise

        return matrix



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements the scheduler classes and functions. This includes the SchedulerBuilder class to create the
general scheduler objects as well as the LRFinderScheduler the custom scheduler for finding the learning rate. The main
difference between the general scheduler and the LRFinder is how they've been used. The LRFinderScheduler is merely used
for finding the learning rates. Unlike other scheduler, the step method is LRFinder called after each bach and not after
each epoch."""
import sys
from copy import deepcopy
import logging

from torch.optim.lr_scheduler import StepLR, MultiStepLR, CosineAnnealingLR, ExponentialLR, ReduceLROnPlateau, CyclicLR, CosineAnnealingWarmRestarts

from learner.data_worker.data_processor import delete_keys_from_dict


class SchedulerBuilder:
    """The class to functionally build the scheduler objects depending on the user input. The main method here is
    build_scheduler. This method uses the type of the scheduler and the parameters to instantiate the corresponding
    object.
    """
    def __init__(self, scheduler_dict):
        """Initialize a SchedulerBuilder object using the scheduler_dict. This dictionary contains all the parameters
        as well as the type of scheduler. The type is the value of the "type" key.

        :param scheduler_dict: the dictionary that contains the type of the scheduler as well as the necessary parameters.
        """
        self._scheduler_type = scheduler_dict["type"] if scheduler_dict else None
        self._scheduler_dict = deepcopy(scheduler_dict)

    @property
    def scheduler_type(self):
        return self._scheduler_type

    @property
    def scheduler_dict(self):
        return self._scheduler_dict

    def build_scheduler(self, optimizer):
        """The main method to get the scheduler object using functional programming. Here, we use the scheduler type to
        get the scheduler object.

        :param optimizer: an optimizer object
        :return: the scheduler object ready to be used during the training.
        """
        if self.scheduler_dict:
            delete_keys_from_dict(self.scheduler_dict, ["type"])
            try:
                return getattr(sys.modules[__name__], self.scheduler_type)(optimizer, **self.scheduler_dict)
            except Exception as e:
                logging.critical(f"Unable to build the scheduler. The error is: {e}")
                sys.exit(1)


class LRFinderScheduler:
    """The scheduler to use when finding the learning rate. This uses a similar logic as the fastai library. The
    original paper can be found here: https://arxiv.org/abs/1506.01186. Here's a useful blog about fastai
    implementation: https://sgugger.github.io/how-do-you-find-a-good-learning-rate.html"""
    def __init__(self, optimizer, start_lr, end_lr, num_it):
        """Initialize a LRFinderScheduler using the input parameters. The majority of the input parameters are defined
        in the configuration file by the user.

        :param optimizer: an optimizer object. We will use this object to update the learning rate.
        :param start_lr: the starting point for the learning rate probe.
        :param end_lr: the end point for the learning rate probe.
        :param num_it: the number of iteration when probing the learning rate. We exponentially go from start_lr to end_lr in this many itertions
        """
        self._optimizer = optimizer
        self._start_lr = start_lr
        self._end_lr = end_lr
        self._num_it = num_it
        self.it = 0

    @property
    def optimizer(self):
        return self._optimizer

    @property
    def start_lr(self):
        return self._start_lr

    @property
    def end_lr(self):
        return self._end_lr

    @property
    def num_it(self):
        return self._num_it

    def step(self):
        """Take one step to update the learning rate. In this schedule we update the learning rate after each batch. We
        first calculate the learning rate depending on the iteration we are at. We will then update the learning rate of
        the optimizer.

        :return: None
        """
        self.it += 1
        lr = self.start_lr * (self.end_lr / self.start_lr) ** (self.it/self.num_it)
        for group in self.optimizer.param_groups:
            group["lr"] = lr

    @property
    def is_done(self):
        """Determine if we have reached the end of the scheduler. This happens when are at or above the number of
        iterations we need.

        :return: true is we've reached the end, false otherwise
        """
        return self.it >= self.num_it



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements shap analysis using a trained model and the test data. It handles everything from getting,
sorting, and writing the values as well as creating the narratives and possibly plots. The engines commonly do the
instantiation and invocation of the classes and methods.
"""
import pandas as pd
import numpy as np
import warnings
import logging
import shap
from shap.plots._beeswarm import pl

from learner.analysis.plot import Plot
from learner.utilities.timer import timeit
from learner.configuration.configuration import Configuration
from learner.data_worker.output_handler import OutputHandler
from learner.data_worker.data_processor import DataProcessor, parrallelize_dataframe, parrallelize_dataframes
from learner.utilities.templates import SHAP_PLOT, SHAP_PLOT_ClASS, SHAP_PATH


class Shap:
    def __init__(self, models_dict_item, conf: Configuration, processor):
        self._models_dict_item = models_dict_item
        self._conf = conf
        # once populated, this dictionary will map each column name to its average
        self.mean_dict = None
        # this temporarily holds class_name for multiprocessing purposes where we can't pass an argument to a method
        self.class_name = None
        self._processor = processor
        # the list of columns that will be created after sorting the shap values. These are then used for creating the
        # narratives
        self.value_feature_col_list = self.get_value_feature_col_list()

    @property
    def conf(self):
        return self._conf

    @property
    def models_dict_item(self):
        return self._models_dict_item

    @models_dict_item.setter
    def models_dict_item(self, value):
        self._models_dict_item = value

    @property
    def processor(self):
        return self._processor

    def get_value_feature_col_list(self):
        """Get a list of column names that would be added to shap output file during the calculation of sorted shape
        values and features

        :return: a list of column name
        """
        value_feature_col_list = list()
        for i in range(self.conf.analysis.shap_num_features):
            pos_val_name = f"positive_value_{i}"
            pos_feat_name = f"positive_feature_{i}"
            neg_val_name = f"negative_value_{i}"
            neg_feat_name = f"negative_feature_{i}"
            value_feature_col_list.extend([pos_val_name, pos_feat_name, neg_val_name, neg_feat_name])
        return value_feature_col_list

    @timeit("run shap analysis")
    def run_shap(self, test_data, tag):
        """The main driver method for performing shap analysis calculations. This is the only method that is usually
        called after instantiating the Shap object.

        :param test_data: the test dataset. This data set contains the training data as well as the id columns.
        :param tag: the model tag. This is being used to name the files.
        is not being used.
        :return: None
        """
        try:
            explainer = shap.TreeExplainer(self.models_dict_item["model"])
        except Exception:
            warnings.warn("The {model_type} model is not supported for shap analysis yet. Learner will continue with "
                          "the rest of calculations...".format(model_type=self.models_dict_item["type"]), UserWarning)
            return

        logging.info("Getting the shap values...")
        logging.info(f"The expected value is {explainer.expected_value}")
        shap_values = explainer.shap_values(test_data[self.models_dict_item["train_features"]])
        # we handle the plots here because in some situations we need only on item in shap_values list and in
        # some situations we need multiple items. Also, we need to plot things before converting shap_values into
        # numpy arrays
        self.handle_plots(tag, shap_values, test_data)
        shap_values = np.array(shap_values)

        # we need the means to use them for storytelling
        self.calculate_the_mean(data=test_data, cols=self.models_dict_item["train_features"])

        if len(shap_values.shape) == 3:
            logging.info("It looks like a model with multiple output was run. Learner will process shap values for "
                         "each output")
            for i, class_name in enumerate(self.models_dict_item["model"].classes_):
                logging.info(f"Processing shap values for class: {class_name}")
                self.process_shap_data(tag, shap_values[i], test_data, class_name=class_name)
        else:
            self.process_shap_data(tag, shap_values, test_data)

    def handle_plots(self, tag, shap_values, test_data):
        """The main method for using shap library and plotting the data. This method checks if plotting functionality
        was activated or not. If it was activated, it continues with the rest of operations.

        :param tag: the model tag. This is being used to name the files.
        :param shap_values: shap values with the same format that comes out of shap library
        :param test_data: the test dataframe
        :return: None
        """
        if not self.conf.analysis.shap_plot_activate:
            logging.info("The shap plotting was not activated, will not plot the data...")
            return

        logging.info("Plotting the shap values...")

        # get the class names if they exist, otherwise set it to None
        try:
            class_names = self.models_dict_item["model"].classes_
        except AttributeError:
            class_names = None

        # try to make bot bar and dot plots
        for plot_type in ["bar", "dot"]:
            logging.info("Making {plot_type} plot for shap values...".format(plot_type=plot_type))

            filename = SHAP_PLOT.format(
                path=self._conf.model.models_dict[tag]["path"],
                output_name=self.conf.workspace.name,
                tag=str(tag),
                type=plot_type,
                sep_timetag=str(self.conf.sep_timetag))
            try:
                # this can fail for dot plot and multi-class classifications problem. In that case, we still want to make
                # the plots but we need to send the data for each class separately.
                shap.summary_plot(shap_values,
                                  test_data[self.models_dict_item["train_features"]],
                                  show=False,
                                  max_display=self.conf.analysis.shap_num_features,
                                  plot_type=plot_type,
                                  class_names=class_names)
                Plot.savefig(filename, pl)
                pl.clf()
            except AssertionError:
                for i, class_name in enumerate(self.models_dict_item["model"].classes_):
                    logging.info("Plotting shap values for class: {c}".format(c=class_name))
                    filename = SHAP_PLOT_ClASS.format(
                        path=self._conf.model.models_dict[tag]["path"],
                        output_name=self.conf.workspace.name,
                        tag=str(tag),
                        type=plot_type,
                        class_name=class_name,
                        sep_timetag=str(self.conf.sep_timetag))
                    shap.summary_plot(shap_values[i],
                                      test_data[self.models_dict_item["train_features"]],
                                      show=False,
                                      max_display=self.conf.analysis.shap_num_features,
                                      plot_type=plot_type,
                                      class_names=class_names)

                    Plot.savefig(filename, pl)
                    pl.clf()

    @timeit("process the shap data")
    def process_shap_data(self, tag, shap_values, test_data, class_name=None):
        """The main method for processing the shap data. This include converting the shap_values numpy arrays in pandas
        dataframe, calling the narrative method to sort the shap values and create the narratives, and saving the final
        data to disk (by calling the write_shap method)

        :param tag: the model tag. This is being used to name the files.
        :param shap_values: a numpy array that contain shap values
        :param test_data: the test dataframe
        :param class_name: the label of the class for classification problems
        :return: None
        """
        logging.info("Processing the shap data...")

        shap_df = pd.DataFrame(data=shap_values, columns=self.models_dict_item["train_features"])
        if self.conf.analysis.shap_include_sorted_values or self.conf.analysis.shap_narrative_activate:
            logging.info("Obtaining the important features and the corresponding shap values for each data point...")
            shap_df = parrallelize_dataframe(shap_df, self.handle_feature_importance, self.conf.analysis.shap_num_cores)
            # this is the serial code
            # shap_df = self.handle_feature_importance(shap_df)

        if not self.conf.analysis.shap_narrative_activate:
            logging.info("The shap narrative functionality was not activated, learner will not generate stories")
        else:
            logging.info("Creating narratives for each data point...")
            self.class_name = class_name

            shap_df, test_data = parrallelize_dataframes(shap_df, test_data, self.handle_narrative, self.conf.analysis.shap_num_cores)
            # this is the serial code
            # self.create_narrative(shap_df, test_data[self.models_dict_item["train_features"]], class_name=class_name if class_name is not None else "target")
        self.write_shap(tag, shap_df, test_data[self.conf.column.id_cols], class_name="_{c}".format(c=class_name) if class_name is not None else '')

    @timeit("compute the mean of the columns")
    def calculate_the_mean(self, data, cols):
        """Accept a dataframe and a list of column. If use_training_mean was set to true, get the mean_dict
        (a dictionary in which the key is the column name and the value is the average value of that column), otherwise
        compute the mean of each column and save them to populate mean_dict.

        :param data: a pandas dataframe
        :param cols: a list of columns to do the calculations
        :return: None
        """
        # we don't need to calculate the mean if narrative is not active
        if not self.conf.analysis.shap_narrative_activate:
            return
        if self.conf.analysis.shap_use_training_mean:
            logging.info("Using the mean of features from the training data to use in narratives.")
            self.mean_dict = self.processor.mean_dict
        else:
            logging.info("Calculating the mean of each feature in test data to use in narratives...")
            processor = DataProcessor(self.conf)
            processor.handle_mean(data[cols])
            self.mean_dict = processor.mean_dict

    def handle_feature_importance(self, shap_df):
        """Use apply method to make a call to _get_important_features method to obtain the important features along with
        their corresponding shap values.

        :param shap_df: the shap dataframe
        :return: updated shap dataframe
        """
        num_features = min(self.conf.analysis.shap_num_features, shap_df.shape[1])
        for i in range(num_features):
            shap_df[f"positive_value_{i}"] = -np.inf
            shap_df[f"positive_feature_{i}"] = -np.inf

            shap_df[f"negative_value_{i}"] = -np.inf
            shap_df[f"negative_feature_{i}"] = -np.inf

        for index, row in shap_df.iterrows():
            sorted_row = row.sort_values(inplace=False, ascending=False)
            for i in range(num_features):
                shap_df.loc[index, f"positive_value_{i}"] = sorted_row.iloc[i] if sorted_row.iloc[i] > 0 else None
                shap_df.loc[index, f"positive_feature_{i}"] = sorted_row.index[i] if sorted_row.iloc[i] > 0 else None

                # 4*num_features is because of the new columns we just created, which will go to the end because of
                # -np.inf
                shap_df.loc[index, f"negative_value_{i}"] = sorted_row.iloc[-1 - i - 4*num_features] if sorted_row.iloc[-1 - i - 4*num_features] < 0 else None
                shap_df.loc[index, f"negative_feature_{i}"] = sorted_row.index[-1 - i - 4*num_features] if sorted_row.iloc[-1 - i - 4*num_features] < 0 else None

        return shap_df

    def handle_narrative(self, shap_df, test_data):
        """Use shap_df and test_data to create the narratives. This method calls create_narrative method to do the work.

        :param shap_df: the shap data frame
        :param test_data: the test dataframe
        :return: updated shap and test_data
        """
        self.create_narrative(shap_df, test_data[self.models_dict_item["train_features"]], class_name=self.class_name if self.class_name is not None else "target")
        return shap_df, test_data

    def create_narrative(self, shap_df, test_df, class_name):
        """The main method that drives generation of the narratives. This method extract the templates and the generic
        messages, and calls narrative_analysis method to add the narrative to the shap data frame.

        :param shap_df: the shap dataframe
        :param test_df: the test dataframe
        :param class_name: the name of the class if we have a multiclass classification problem, or "target" otherwise
        :return: None
        """
        try:
            templates = self.conf.analysis.shap_narrative_stories[class_name]["templates"]
            generic = self.conf.analysis.shap_narrative_stories[class_name]["generic"]
            self._narrative_analysis(shap_df, test_df[self.models_dict_item["train_features"]], templates, generic)
        except KeyError:
            # Note: this warning may not show in the log file due to the mechanism for multiprocessing
            warnings.warn("No narrative parameters for the class {class_name} was provided. Learner will create generic "
                          "narrative for this class. Please update your configuration file if you'd like to create custom "
                          "narratives.".format(class_name=class_name), Warning)
            self._narrative_analysis(shap_df, test_df[self.models_dict_item["train_features"]], [], None)

    def _narrative_analysis(self, shap_df, test_df, templates, generic):
        """Iterate through shap dataframe and test dataframe concurrently and make a call to _get_narrative method to
        get the text for the narrative. Then add that narrative to shap dataframe

        :param shap_df: the shap dataframe with all sorted and unsorted features (we don't use the unsorted columns)
        :param test_df: the test dataframe that include all features. This dataframe is being used to pull the actual numbers
        :param templates: the list of templates. The first matched item will be used from this list
        :param generic: the generic narrative message
        :return: None
        """
        for (i, shap_row), (_, test_row) in zip(shap_df.iterrows(), test_df.iterrows()):
            narrative = self._get_narrative(shap_row, test_row, templates, generic)
            shap_df.loc[i, "narrative"] = narrative

    def _get_narrative(self, shap_row, test_row, templates, generic):
        """Get the narrative text. To do this, we first get the positive and negative features. Knowing the positive and negative features,
        we go through the templates one by one. If that template is a match, we replace the templates with appropriate
        values and return the narrative. If there's no match we return the generic narrative.

        :param shap_row: a row in the shap dataframe
        :param test_row: a row in the test dataset
        :param templates: the list of templates provided in the configuration file
        :param generic: the generic narrative. Currently, we just return it but in the future we may want to get more creative with it.
        :return: the narrative - custom or generic
        """
        # we take care of exclude_cols here
        positive_features = {}
        for i, feature in enumerate(shap_row.index):
            if (str(feature).startswith("positive_feature_") and
                    shap_row.iloc[i] and
                    shap_row.iloc[i] not in self.conf.analysis.shap_exclude_cols and
                    not isinstance(shap_row.iloc[i], float)):

                feature_index = feature.rsplit("_")[-1]
                positive_features[shap_row.iloc[i]] = shap_row["positive_value_{i}".format(i=feature_index)]

        negative_features = {}
        for i, feature in enumerate(shap_row.index):
            if (str(feature).startswith("negative_feature_") and
                    shap_row.iloc[i] and
                    shap_row.iloc[i] not in self.conf.analysis.shap_exclude_cols and
                    not isinstance(shap_row.iloc[i], float)):

                feature_index = feature.rsplit("_")[-1]
                negative_features[shap_row.iloc[i]] = shap_row["negative_value_{i}".format(i=feature_index)]

        # this list will hold all the narratives for the importance method
        importance_narratives = []
        for template in templates:
            if self._is_match(template, positive_features.keys(), negative_features.keys(), test_row):
                if self.conf.analysis.shap_narrative_method == "order":
                    kwargs = self._construct_kwargs(template, test_row)
                    narrative = template["narrative"].format(**kwargs)
                    return narrative
                elif self.conf.analysis.shap_narrative_method == "importance":
                    score = self._get_match_score(template, positive_features, negative_features)
                    importance_narratives.append((template, score))

        # if we found a match for "importance" method
        if importance_narratives:
            template = sorted(importance_narratives, key=lambda x: x[1])[-1][0]
            kwargs = self._construct_kwargs(template, test_row)
            narrative = template["narrative"].format(**kwargs)
            return narrative

        # if no generic provided, we build a generic narrative using data
        if not generic:
            return self._build_generic_narrative_with_data(test_row, positive_features, negative_features)
        # if a generic narrative was provided and we didn't have a match, we return that generic narrative
        return generic

    @staticmethod
    def _get_match_score(template, positive_features, negative_features):
        """For the narratives that there's a match, compute the match score but adding the absolute shap values
        together.

        :param template: an item in the templates list provided in the configuration file.
        :param positive_features: a dictionary in which the keys are features with the highest positive values and the
          values are the shap values for those features
        :param negative_features: a dictionary in which the keys are features with the highest negative values and the
          values are the shap values for those features
        :return: the score of the matched narrative
        """
        score = 0
        for feature, value in template["features"].items():
            try:
                score += abs(positive_features[feature])
            except KeyError:
                score += abs(negative_features[feature])

        return score

    @staticmethod
    def _is_match(template, positive_features, negative_features, test_row):
        """Accept a template item and go through all the features in the template. If all the features, their shap
        category (positive or negative) and their values (if defined) matches the one predicted by the model,
        then return true. If there's any mismatched, return false.

        :param template: an item in the templates list provided in the configuration file.
        :param positive_features: a dictionary in which the keys are features with the highest positive values and the
          values are the shap values for those features
        :param negative_features: a dictionary in which the keys are features with the highest negative values and the
          values are the shap values for those features
        :param test_row: a row in the test dataset
        :return: a boolean indicating if we have a match or not
        """
        while True:
            for feature, value in template["features"].items():
                if value["impact"] == "positive" and feature not in positive_features:
                    return False
                if value["impact"] == "negative" and feature not in negative_features:
                    return False
                # if values are defined but the value for the feature does not exist in the list, return false
                if value["values"] and test_row.loc[feature] not in value["values"]:
                    return False
            return True

    def _construct_kwargs(self, template, test_row):
        """Create a kwargs dictionary to be used for replacing the templates with actual values. Currently, each
        templates accepts templates corresponding to the features in the "features" section of the template item.
        We need to replace each template with their actual values.

        :param template: a template item in templates list
        :param test_row: a row in the test dataframe
        :return: a dictionary of the keyword arguments
        """
        kwargs = {}
        for feature, value in template["features"].items():
            if value["include_mean"]:
                kwargs[feature] = "{value} (avg: {average:.3f})".format(value=test_row.loc[feature], average=self.mean_dict[feature])
            else:
                kwargs[feature] = "{value}".format(value=test_row.loc[feature])
        return kwargs

    def _build_generic_narrative_with_data(self, test_row, positive_features, negative_features):
        """Build a generic narrative based on the sign of the shap values. Basically, the narrative says what features
        are driving the prediction up (with positive shap values) and what features are driving the predictions down
        (with negative shap values). It prints out the name of those features along with their value and the average
        of the cohort, i.e. the test data we are making the predictions for.

        :param test_row: a row in the test dataset
        :param positive_features: a list of features with positive shap values
        :param negative_features: a list of features with negative shap values
        :return: the text of the narrative that says which features driving the predictions up or down
        """
        narrative = []
        if positive_features:
            narrative.append("These features are driving the predictions up:")
            for feature in positive_features:
                narrative.append(feature)
                narrative.append("(val: {value}, avg: {average:.3f}),".format(value=test_row[feature],
                                                                        average=self.mean_dict[feature]))

        if negative_features:
            narrative.append("\nThese features are driving the predictions down:")
            for feature in negative_features:
                narrative.append(feature)
                narrative.append("(val: {value}, avg: {average:.3f}),".format(value=test_row[feature],
                                                                            average=self.mean_dict[feature]))

        return " ".join(narrative)

    @timeit("write the shap output")
    def write_shap(self, tag, shap_df, data, class_name=''):
        """Write the shap dataframe along with any additional columns (typically id columns) to disk. This method also
        makes a call to drop_cols method of the DataProcessor class to drop the unsorted column if not requested.

        :param tag: the model tag defined by user in the configuration file.
        :param shap_df: the shap dataframe
        :param data: the additional data to be concatenated with the shap dataframe. This is typically id columns
        :param class_name: the class name in case of multi-class classification problem.
        :return: None
        """
        logging.info("Writing the shap output data...")
        # we need to drop the unsorted columns if the user hasn't asked for it
        if not self.conf.analysis.shap_include_raw_values:
            DataProcessor.drop_cols(shap_df, list(self.models_dict_item["train_features"]))
        if not self.conf.analysis.shap_include_sorted_values:
            DataProcessor.drop_cols(shap_df, self.value_feature_col_list, errors="ignore")

        data.reset_index(drop=True, inplace=True)
        shap_df.reset_index(drop=True, inplace=True)
        results = pd.concat([data, shap_df], axis=1, ignore_index=False)

        filename = SHAP_PATH.format(
            output_name=self._conf.workspace.name,
            tag=str(tag),
            class_name=class_name,
            sep_timetag=str(self._conf.sep_timetag))

        output = OutputHandler(self._conf)
        output.save_file(self._conf.model.models_dict[tag]["path"], filename, results, add_timetag=False)
        logging.info("Successfully wrote the shap output data")



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for processing image data. Currently, this modules implement a ImageTransformer class. This class
builds a list of transforms to be passed to PyTorch's Compose class. This module is currently being used by the
image_classifier engine."""

from copy import deepcopy
import logging

from torchvision import transforms

from learner.data_worker.data_processor import delete_keys_from_dict


class ImageTransformer:
    """The ImageTransformer builds a list of transformation and passes that to the PyTorch's Compose class to create
    a transform object. That object is then passed to Dataset classes. The main method is build_transform.
    """
    def __init__(self, conf, data_type):
        """Initialize a ImageTransformer using a conf object and the data_type.

        :param conf: a conf object
        :param data_type: the data_type, this could be train, validation, or test.
        """
        self._conf = conf
        self._data_type = data_type
        self.keys_to_delete = ["type", "activate"]

    @property
    def conf(self):
        return self._conf

    @property
    def data_type(self):
        return self._data_type

    def build_transform(self):
        """The main method for building the transform list. This method uses the input provided by the user to create
        the necessary transform objects. Each method is aware of data_type value. The methods would return None if they
        should not create an object for a data_type value. That's why we check the value returned from the method
        before appending it to the list.

        :return: a Compose object to be passed to the Dataset classes
        """
        logging.info("Building image transformations...")
        transform_list = []
        if self._conf.process.image_transform_params:
            params = deepcopy(self._conf.process.image_transform_params)
            for param in params:
                transform = getattr(self, f"build_{param['type'].lower()}")(param)
                if transform:
                    transform_list.append(transform)
        return transforms.Compose(transform_list)

    def build_randomrotation(self, transform):
        if self._data_type != "train":
            return
        delete_keys_from_dict(transform, self.keys_to_delete)
        return transforms.RandomRotation(**transform)

    def build_randomhorizontalflip(self, transform):
        if self._data_type != "train":
            return
        delete_keys_from_dict(transform, self.keys_to_delete)
        return transforms.RandomHorizontalFlip(**transform)

    def build_resize(self, transform):
        delete_keys_from_dict(transform, self.keys_to_delete)
        return transforms.Resize(**transform)

    def build_centercrop(self, transform):
        delete_keys_from_dict(transform, self.keys_to_delete)
        return transforms.CenterCrop(**transform)

    def build_normalize(self, transform):
        delete_keys_from_dict(transform, self.keys_to_delete)
        return transforms.Normalize(**transform)

    def build_totensor(self, transform):
        return transforms.ToTensor()





# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for handling the communications. This module instantiates email and other communication classes and
calls their driver method to perform the communications."""
from learner.configuration.configuration import Configuration


class CommunicationHandler:
    """Instantiate appropriate objects and call the communicator methods to handle the communications"""

    def __init__(self, conf: Configuration, data=None, message=None):
        """Accept a conf object, data, and message to instantiate the object.

        :param conf: an instance of the Configuration class
        :param data: the data to use for communication purposes
        :param message: the message (typically text) to use for the communication
        """
        self._conf = conf
        self._data = data
        self._message = message

    @property
    def conf(self):
        return self._conf

    @property
    def data(self):
        return self._data

    @property
    def message(self):
        return self._message

    def handle_admin_communication(self):
        """Depending on the parameters, import the relevant class, instantiate the object and call the drive method.

        :return: None
        """
        if self._conf.communication.admin_email_activate:
            from learner.communication.email_manager import AdminEmailManager
            communication_manager = AdminEmailManager(self._conf, message=self.message, data=None)
            communication_manager.handle_communications()



#! /usr/bin/env python
import os
from setuptools import setup, find_packages

# to run the module do the following command:
# python setup.py sdist


def get_dependencies():
    """This function reads the requirements.txt file and creates a list of packages to be installed before Learner can
    be installed. In practice, we should be defining a separate dependencies in the setup because the requirements file
    is too restrictive but this is fine for now. As long as the users use docker containers or virtual environments,
    this should't cause any issues.

    :return: a list of libraries/packages to install when installing Learner.
    """
    with open('requirements.txt') as f:
        required = f.read().splitlines()
    return required


def compile_learner():
    """Go through each package and modules, and compile them.

    :return: None
    """
    setup(
        name='clearner',
        version='0.9.75',
        description="Learner is a software platform for building production-ready machine learning models without writing any codes.",
        author="Prizmi LLC",
        author_email="contact@prizmi.ai",
        python_requires='>=3.7,<3.8',
        url="https://prizmi.ai/learner/home",
        license="Other",
        build_dir="build",
        install_requires=get_dependencies(),
        packages=find_packages(exclude=("tests.*", "tests", "docs.*", "docs")) + ['.'],
        package_data={'learner.schema': ['*.json'],
                      ".": ['requirements.txt', 'LICENSE'],
                      },
        entry_points={
            'console_scripts': ['learner=learner.setup.main:main'],
        },
    )


def clean_up():
    os.system('rm -r *.egg-info')


if __name__ == "__main__":
    compile_learner()
    clean_up()
